import sys
import math
import Orange
import orange
import cStringIO

from collections import defaultdict
from scorpion.util.rangeutil import *


inf = infinity = 1e10000

#class SDRule (orange.Rule):
class SDRule(object) :
  __id__ = 0

  
  def __init__(self, data, targetClass, conditions=[] ,g =1, negate=False, c_range = None):
      self.g = g
      self.data = data
      self.targetClass = targetClass
      self.filter = orange.Filter_values(domain = data.domain,
                                          conditions=conditions,
                                          conjunction=1,
                                          negate=negate)
      self.__examples__ = None
      self.filterAndStore()   # set examples, classifier, distribution; TP, calculate quality, complexity, support
      self.id = SDRule.__id__
      self.fixed = False # fixed rules cannot be extended

      self.weight = 1.
      self.quality = -inf
      self.score = 0.
      self.score_norm = None
      self.inf_state = None
      self.isbest = False

      self.c_range = c_range
      if not self.c_range: self.c_range = [infinity, -infinity]

      self.stats_mean = None
      self.stats_std = None
      self.stats_meannorm = None
      self.stats_nmean = None
      self.stats_nstd = None
      self.stats_nmeannorm = None
      self.stats_max = None
      self.parent_rule = None

      # rules within the cluster (for Basic.group_rules)
      self.cluster_rules = set()
      SDRule.__id__ += 1

  def __improvement__(self):
      if self.parent_rule:
          return self.quality - self.parent_rule.quality
      return self.quality
  improvement = property(__improvement__)

  def __get_examples__(self):
      if not self.__examples__:
          self.__examples__ = self.filter_table(self.data)
      return self.__examples__
  examples = property(__get_examples__)

  def filterAndStore(self):
      self.complexity = len(self.filter.conditions)
      return
      c = 0
      for cond in self.filter.conditions:
          c += 1
          if type(cond) == Orange.data.filter.ValueFilterDiscrete:
              c += max(0., 0.2 - 1. / len(cond.values))
      
      self.complexity = c

  def covers(self, example):
      return len(self.filter([example])) > 0

  def filter_table(self, table, **kwargs):
      try:
          return table.filter_ref(self.filter, **kwargs)
      except:
          return self.filter(table, **kwargs)

  def __call__(self, table):
      return self.filter_table(table)


  def fill_in_rule(self, table, ref_bounds):
    domain = table.domain

    # if there are any cols not in the rule, fill them in with table bounds
    conds = {}
    for c in self.filter.conditions:
      attr = domain[c.position]
      name = attr.name
      conds[name] = True

    for col, bounds in ref_bounds.iteritems():
      if col in conds:
          continue

      attr = domain[col]
      pos = domain.index(attr)

      if bounds is None:
        vals = range(len(attr.values))
        vals = [orange.Value(attr, attr.values[v]) for v in vals]
        cond = orange.ValueFilter_discrete(position=pos, values=vals)
      else:
        (minv, maxv) = bounds
        
        cond = orange.ValueFilter_continuous(
          position=pos,
          oper = orange.ValueFilter.Between,
          min = minv-1,
          max = maxv+1
        )
      self.filter.conditions.append(cond)


  def cloneAndAppendRule(self, rule):
      conds = list(self.filter.conditions)
      conds.extend(rule.filter.conditions)

      return SDRule(self.data, self.targetClass, conds, self.g)

  def cloneAndAppendCondition(self, cond):
    attr = self.data.domain[cond.position]
    if type(cond) == Orange.data.filter.ValueFilterContinuous:
      return self.cloneAndAddContCondition(attr, cond.min, cond.max)
    else:
      values = cond.values
      vals = [attr.values[int(v)] for v in values]
      return self.cloneAndAddCondition(attr, vals)



  def cloneAndAddCondition(self, attribute, values, used=False, negate=False):
      '''Returns a copy of this rule which condition part is extended by attribute = value'''
      conds = list(self.filter.conditions)

      if not(values):
          return self
      if not isinstance(values, list):
          values = [values]

      pos = self.data.domain.index(attribute)                    
      conds = filter(lambda cond: cond.position != pos, conds)
      values = [orange.Value(attribute, value) for value in values]
      conds.append(
          orange.ValueFilter_discrete(
              position = self.data.domain.index(attribute),
              values = values
              )
          )
      conds.sort(key=lambda c: c.position)

      return SDRule (self.data, self.targetClass, conds, self.g)

  def cloneAndAddContCondition(self, attribute, minv, maxv, op=orange.ValueFilter.Between, used=False, negate=False):
      conds = list(self.filter.conditions)
      pos = self.data.domain.index(attribute)

      if used:
          for cond in conds:
              if cond.position == pos:
                  assert op == orange.ValueFilter.Between
                  minv = max(cond.min, minv)
                  maxv = min(cond.max, maxv)
          conds = filter(lambda cond: cond.position != pos, conds)
      vfc = Orange.data.filter.ValueFilterContinuous(
          position=pos,
          oper=op,
          min=minv,
          max=maxv)

      conds.append( vfc )
      conds.sort(key=lambda c: c.position)
      return SDRule (self.data, self.targetClass, conds, self.g)

  def cloneAndNegate(self, negate=True):
      conds = self.filter.conditions[:]
      return SDRule(self.data, self.targetClass, conds, self.g, negate=negate)

  def clone(self, data=None):
      if not data or data == self.data:
        ret = self.cloneAndNegate(self.filter.negate)
        ret.__examples__ = self.__examples__
      else:
        ret = SDRule(data, self.targetClass, self.filter.conditions[:], self.g, negate=self.filter.negate)

      ret.cluster_rules = set(self.cluster_rules)
      if self.c_range is not None:
        ret.c_range = list(self.c_range)
      if self.inf_state is not None:
        ret.inf_state = list(self.inf_state)
      ret.quality = self.quality
      return ret

  def cloneWithNewData(self, newdata):
      conds = self.filter.conditions[:]
      rule = SDRule(newdata, self.targetClass, conds, self.g, negate=self.filter.negate)
      rule.quality = self.quality
      rule.fix = self.fixed
      rule.stats_mean = self.stats_mean
      rule.stats_std = self.stats_std 
      rule.stats_nmean = self.stats_nmean
      rule.stats_nstd = self.stats_nstd 
      rule.stats_max = self.stats_max 
      return rule


  def isSubsumed(self, rule):
      """
      Returns True if self is subsumed by rule
      subsumed if self's conditions are tighter than rule, but
      has the same number examples
      """
      for c in rule.filter.conditions:
          found = False
          for myc in self.filter.conditions:
              # does c contain myc?
              if c.position != myc.position:
                  continue
              if rule.data.domain[c.position].varType != self.data.domain[myc.position].varType:
                  continue
              if rule.data.domain[c.position].varType == orange.VarTypes.Discrete:
                  domain = rule.data.domain[c.position]
                  cvals = [domain.values[int(vidx)] for vidx in c.values]
                  
                  domain = self.data.domain[myc.position]
                  mycvals = [domain.values[int(vidx)] for vidx in myc.values]
                  
                  if not set(cvals).issuperset(set(mycvals)):
                      continue
              else:
                  if not(myc.min >= c.min and myc.max <= c.max):
                      continue                    
              found = True
              break
          if not found:
              return False
      return True
      
  
  def isIrrelevant(self, rule):
      '''Returns True if self is irrelevant compared to rule.'''
      def isSubset(subset, set):
          if len(subset) > len(set):
              return False
          else:
              index = 0;
              for e in subset:
                  while index<len(set) and e != set[index]:
                      index+=1
                  if index >= len(set):
                      return False
              return True

      if isSubset(self.TP, rule.TP) and isSubset(rule.FP, self.FP):
          return True
      else:
          return False

  def __lt__(self, o):
      if self.quality < o.quality:
          return True
      elif self.quality == o.quality:
          if self.complexity > o.complexity:
              return True
          elif self.complexity == o.complexity:
              if len(self.examples) < len(o.examples):
                  return True
      return False

  def __gt__(self, o):
      return o.__lt__(self)

  def eq(self, o):
      return (self.quality == o.quality and
              self.complexity == o.complexity and
              len(self.examples) == len(o.examples))

  def __ne__(self, o):
      return not self.eq(o)
      
  def __cmp__(self, other):
      if self.quality < other.quality:
          return -1
      if self.quality > other.quality:
          return 1
      if self.complexity > other.complexity:
          return -1
      if self.complexity < other.complexity:
          return 1
      if len(self.examples) < len(other.examples):
          return -1
      if len(self.examples) > len(other.examples):
          return 1
      return 0

  def __hash__(self):
      condStrs = self.toCondStrs()
      condStrs.append('rule')
      condStrs.sort()
      return hash(tuple(condStrs))


  def __eq__(self, o):
      return hash(self) == hash(o)

  def __str__(self):
      return self.ruleToString()

  def __attributes__(self):
      domain = self.data.domain
      attrs = list(set([domain[c.position].name for c in self.filter.conditions]))
      attrs.sort()
      return tuple(attrs)
  attributes = property(__attributes__)


  def simplify(self, data=None, cdists=None, ddists=None):
    """
    Args:
      data:   non-filtered! data
      cdists: non-filtered Continuous distribution
      ddists: non-filtered discrete distribution
    Return:
      copy of this rule with simplified conditions
    """
    subset = data and self(data) or self.examples
    data = data or self.data
    ret = self.clone()

    positions = [cond.position for cond in self.filter.conditions]
    cdists = cdists or Orange.statistics.basic.Domain(data)
    ddists = ddists or Orange.statistics.distribution.Domain(data)
    #scdists = Orange.statistics.basic.Domain(subset)
    #sddists = Orange.statistics.distribution.Domain(subset)

    conds = []
    for old_cond, idx in zip(self.filter.conditions, positions):
      attr = data.domain[idx]

      # if rule values == full dataset values, then remove rule
      # filter down to the values that intersect the subset of data
      if attr.var_type == Orange.feature.Type.Discrete:
        full_d = ddists[attr.name]
        #sub_d = sddists[attr.name]
        fvals = [k for k,v in full_d.items() if v]
        cvals = set([str(attr.values[int(v)]) for v in old_cond.values])
        if len(cvals) == len(fvals):
          continue

        #dvals = [k for k,v in sub_d.items() if v]
        #vals = set(cvals).intersection(dvals)
        vals = cvals
        cond = orange.ValueFilter_discrete(
          position = idx,
          values = [orange.Value(attr, val) for val in vals]
        )
      else:
        fb = cdists[attr.name]
        #sb = scdists[attr.name]
        old_bound = [fb.min, fb.max]
        cond_bound = [old_cond.min, old_cond.max]

        bound = r_intersect(old_bound, cond_bound)
        if r_vol(bound) >= r_vol(old_bound): continue
        #bound = r_intersect(bound, [sb.min, sb.max])
        cond = old_cond
        cond.min, cond.max = bound[0], bound[1]
      conds.append(cond)
      continue



    ret.quality = self.quality
    ret.filter.conditions = conds
    ret.c_range = list(self.c_range)
    return ret



  def condToString(self, c):
      domain = self.data.domain
      name = domain[c.position].name        

      if domain[c.position].varType == orange.VarTypes.Discrete:
          if len(c.values) == 0:
              return
          if len(c.values) == 1:
              vidx = int(c.values[0])
              v = domain[c.position].values[vidx]
          elif len(c.values) == len(domain[c.position].values):
              return
          else:
              vs = [domain[c.position].values[int(vidx)] for vidx in c.values]
              vs = map(str, vs)
              strlen = len(vs[0])
              idx = 0
              while idx < len(vs)-1 and strlen + len(vs[idx+1]) < 60:
                  strlen += len(vs[idx+1])
                  idx += 1
              v = ', '.join(vs[:idx+1])
              if idx < len(vs)-1:
                  v = '%s (%d+ ..)' % (v, len(vs) - 1 - idx)
          return '%s = %s' % (name, v)

      elif domain[c.position].varType == orange.VarTypes.Continuous:
          return '%.7f <= %s < %.7f' % (c.min, name, c.max) 

      return None

  def condToDict(self, c):
    """
    Translate condition into a dictionary necoding the column, data type and values
    Useful for JSONifying the rule
    """
    ret = {
      'pos': None,
      'col': None,
      'type': None,
      'vals': None
    }

    domain = self.data.domain
    ret['col'] = name = domain[c.position].name        
    ret['pos'] = c.position

    if domain[c.position].varType == orange.VarTypes.Discrete:
      if len(c.values) == 0 or len(c.values) == len(domain[c.position].values):
        return None

      def reencode(v):
        if v == 'None' or v == 'NULL':
          return None
        return v
      vals = [domain[c.position].values[int(vidx)] for vidx in c.values]
      vals = map(reencode, vals)
      ret['type'] = 'str'
      ret['vals'] = vals
    else:
      ret['type'] = 'num'

      if any([t in str(type(c.min)) for t in ['date', 'time']]):
        ret['type'] = 'timestamp'
      
      ret['vals'] = [max(-1e100, c.min), min(1e100, c.max)]

    return ret
  
  @staticmethod
  def dictToCond(d, data):
    if d['type'] == 'num':
      return orange.ValueFilter_continuous(
          position = d['pos'],
          oper = orange.ValueFilter.Between,
          min = d['vals'][0],
          max = d['vals'][1]
      )

    # XXX: NULL hack
    attr = data.domain[d['col']]
    vals = []
    for v in d['vals']:
      if v is None:
        if 'NULL' in attr.values:
          v = 'NULL'
        elif 'None' in attr.values:
          v = 'None'
      vals.append(orange.Value(attr, v))
    return orange.ValueFilter_discrete(position=d['pos'], values=vals)


  def ruleToString(self):
    domain = self.data.domain
    ret = []
    for i,c in enumerate(self.filter.conditions):
      s = self.condToString(c)
      if s:
        ret.append(s)

    ret.sort()
    rule = ' and '.join(ret)
    if self.filter.negate:
      rule = '%s (Neg)' % rule
    return '%.2f %d  %s' % ((self.quality or -inf), len(self.examples), rule)

  def toCondStrs(self):
    domain = self.data.domain
    ret = []
    if self.filter.negate:
      ret.append('neg')

    for i,c in enumerate(self.filter.conditions):
      s = self.condToString(c)
      if s:
        ret.append(s)

    ret.sort()
    return ret
  cond_strs = property(toCondStrs)


  def toCondDicts(self):
      domain = self.data.domain
      if self.filter.negate:
        # DON'T SUPPORT NEGATION YET
        pass

      dicts = map(self.condToDict, self.filter.conditions)
      dicts = filter(bool, dicts)
      return dicts
  cond_dicts = property(toCondDicts)

  def to_json(self):
    return {
      #'conds': self.toCondDicts(),
      'clauses': self.cond_dicts,
      'count': len(self.examples),
      'c_range': self.c_range,
      'quality': self.quality,
      'score': self.quality,
      'inf_state': self.inf_state,
      'alt_rules': [r.cond_dicts for r in self.cluster_rules]
    }

  @staticmethod
  def from_json(j, data=None):
    conds = [SDRule.dictToCond(d, data) for d in j['clauses']]
    rule = SDRule(data, None, conds, None)
    rule.c_range = j['c_range']
    rule.quality = j['quality']
    rule.inf_state = j['inf_state']
    return rule


  
  def printRule(self):
      fmt = "quality= %2.2f complexity=%2d covered=%2d %s"
      return  fmt % (self.quality,
                      self.complexity,
                      len(self.examples),
                      self.ruleToString())



  def getUndiscretized(self, original_data):
      cond = []
      for c in self.filter.conditions:
          d_attribute = self.data.domain[c.position]
          if d_attribute in original_data.domain:
              c.position = original_data.domain.index(d_attribute)
              cond.append(c)
          else:
              position = original_data.domain.index(original_data.domain[d_attribute.name]) #[2:]])

              points = d_attribute.getValueFrom.transformer.points
              value_idx = int(c.values[0])

              if value_idx == 0: # '<='
                  cond.append(
                          orange.ValueFilter_continuous(
                                  position = position,
                                  max = points[0],
                                  min = float(-infinity),
                                  outside = False)
                              )
              elif 0 < value_idx < len(points): # (x,y]
                  cond.append(
                          orange.ValueFilter_continuous(
                                  position = position,
                                  max = points[value_idx],
                                  min = points[value_idx-1],     # zaprti interval '[' namesto odprti '('
                                  outside = False)
                              )
              elif value_idx == len(points): # '>'
                  cond.append(
                          orange.ValueFilter_continuous(
                                  position = position,
                                  max = float(infinity),
                                  min = points[-1],
                                  outside = True)
                              )


      rule = SDRule(original_data, self.targetClass, cond, self.g)
      rule.quality = self.quality
      rule.fix = self.fixed
      rule.score = self.score
      rule.stats_mean = self.stats_mean
      rule.stats_std = self.stats_std 
      rule.stats_nmean = self.stats_nmean
      rule.stats_nstd = self.stats_nstd 
      rule.stats_max = self.stats_max
      return rule
              










class SDRules(object):
    def __init__(self, listOfRules, targetClassRule, algorithmName="SD"):
        self.rules = listOfRules
        self.targetClassRule = targetClassRule
        self.algorithmName = algorithmName

    def makeSelection(self, indexes):
        """Returns the rules that are at specified indexes."""
        rulesSubset = []
        for i in indexes:
            rulesSubset.append(self.rules[i])
        return SDRules(rulesSubset, self.targetClassRule)

    def sortByConf(self):
        self.rules.sort(lambda x, y: -cmp(x.confidence, y.confidence))

    def printRules(self):
        for rule in self.rules:
            rs = rule.printRule()
            print rs

def rule_to_json(rule, **kwargs):
  """
  Args
    kwargs: other key-val pairs to add in the json object
            e.g., yalias: "xxx"
  Returns
    JSON object
  """
  def rnd(v):
    if v == float('-inf'): return -1e100
    if v == float('inf'): return 1e100
    return round(v, 3)

  ret = rule.to_json()
  ret['c_range'] = map(rnd, ret['c_range'])
  ret['score'] = rnd(ret['score'])
  ret['quality'] = rnd(ret['quality'])
  return ret


def fill_in_rules(rules, table, cols=None, cont_dists=None):
    # compute bounds for columns in self.cols
    if cols is None:
        cols = [attr.name for attr in table.domain]

    nparr = None
    ref_bounds = {}
    for col in cols:
        attr = table.domain[col]
        if attr.var_type == Orange.feature.Type.Discrete:
            ref_bounds[col] = None
            continue

        if cont_dists:
          bound = cont_dists[attr.name].min, cont_dists[attr.name].max
        else:
          if nparr is None:
            nparr = table.to_numpyMA('ac')[0]
          pos = table.domain.index(attr)
          arr = nparr[:,pos]
          bound = (arr.min(), arr.max())

        ref_bounds[col] = bound


    for rule in rules:
        rule.fill_in_rule(table, ref_bounds)



# model rules as trees.  look for opportunities to prune --> how?



from ..util import *

from clique import MR
from naivegrouper import NaiveGrouper
from basic import Basic

class NaiveMR(MR):
    def setup_tables(self, full_table, bad_tables, good_tables, **kwargs):
        Basic.setup_tables(self, full_table, bad_tables, good_tables, **kwargs)
        self.grouper = NaiveGrouper(full_table, self) 

        self.SCORE_ID = add_meta_column(
                chain([full_table], bad_tables, good_tables),
                'SCOREVAR' 
        )



    def __call__(self, *args, **kwargs):
        self.naive = True

        return MR.__call__(self, *args, **kwargs)





import pdb
import datetime
import operator
import json
import logging
import orange
import Orange
import orngTree
import orngStat
import orngTest
import orngDisc
import Orange.feature as orf
import numpy as np


from datetime import datetime, date, timedelta
from datetime import time as dttime
from collections import deque, defaultdict
from dateutil.parser import parse as dateparse


from scorpionsql.db import *
from scorpionsql.sql import *
from scorpionsql.sqlparser import *
from scorpionsql.aggerror import *

from util import *
from settings import *




class SharedObj(object):

  def __init__(
      self, db,
      errors=[],
      goodkeys={},
      ignore_attrs=[],
      schema=[],
      dbname=None,
      parsed=None,
      params=[],
      **kwargs):
    if not db and not dbname:
        raise "SharedObj requires a database connection!"
    self.db = db or connect(dbname)
    dbname = dbname or str(self.db.url).split("/")[-1]
    self.monetdb = connect(dbname, engine='monet')
    self.dbname = dbname
    self.parsed = parsed
    self.params = params   # parameters for parsed SQL object
    self.errors = errors
    self.goodkeys = goodkeys or {}
    self.schema = schema or db_schema(db, self.parsed.tables[0])
    # aggregate type -> {groupby key -> ids of "bad" tuples}
    self.ignore_attrs = ignore_attrs
    self.merged_tables = {}
    self.rules = {}
    self.top_k_rules = {}
    self.clauses = {}
    self.c = kwargs.get('c', 0.3)

    # created by server to track status of scorpion
    # processing
    # should be set when creating SharedObj
    self.status = None   
    
    if not self.parsed:
      raise Error("expected a parsed SQL object!")

    if len(self.parsed.fr) > 1:
        # XXX: only support single table queries
        raise "Don't support joins yet!"

      
  def clone(self):
    return SharedObj(
      self.db, 
      parsed=self.parsed,
      dbname=self.dbname,
      errors=self.errors,
      goodkeys=self.goodkeys,
      ignore_attrs=self.ignore_attrs,
      schema=self.schema,
      params=self.params
    )

  def get_tuples(self, keys, attrs=None):
    try:
        if keys is None or not len(list(keys)):
            return []
    except:
        pass
    attrs = attrs or self.rules_schema
    return [list(row) for row in self.get_filter_rows(keys=keys, attrs=attrs)]
    

  
  def get_filter_rows(self, keys=None, attrs=None, where=None):
    """
    Need to deal with keys and such outside of function
    """
    qobj = self.parsed.get_filter_qobj(keys=keys)
    if attrs:
        qobj.select = Select(attrs)
    if where:
        qobj.where.append(where)

    params = list(self.params)
    if keys:
      params.append(tuple(list(keys)))

    return query(self.db, str(qobj), [params])


  def get_rules_schema(self):
    """
    """
    invalid_types = [date, datetime, dttime]
    used_attrs = set()
    for selexpr in self.parsed.select:
        used_attrs.update(selexpr.cols)
    
    schema = dict(filter(lambda p: p[1] not in invalid_types, self.schema.iteritems()))
    ret = set(schema.keys()).difference(used_attrs)
    ret.add('id')
    return ret

  def update_status(self, s):
    if self.status:
      self.status.update_status(s)

  def update_rules(self, label, rules):
    if self.status:
      self.status.update_rules(label, rules)

  attrnames = property(lambda self: self.schema.keys())
  rules_schema = property(get_rules_schema)
  sql = property(lambda self: str(self.parsed))
  prettify_sql = property(lambda self: self.parsed.prettify())
  filter = property(lambda self: self.parsed.get_filter_qobj())






def create_sharedobj(dbname, sql, badresults, goodresults, errtype):
  from arch import get_provenance
  db = connect(dbname)
  parsed = parse_sql(sql)
  obj = SharedObj(db, parsed=parsed, dbname=dbname)


  qobj = obj.parsed
  nonagg = qobj.select.nonaggs[0]
  xcol = nonagg.cols[0]
  col_type = db_type(db, qobj.fr, xcol)

  # assumes every aggregate has the same bad keys
  badresults = extract_agg_vals(badresults, col_type)
  goodresults = extract_agg_vals(goodresults)

  errors = []
  for agg in qobj.select.aggregates:
    aggerr = AggErr(agg, badresults, 20, errtype, {'erreq' : None})
    errors.append(aggerr)

    label = agg.shortname
    obj.goodkeys[label] = goodresults
  obj.errors = errors

  table = get_provenance(obj, obj.errors[0].agg.cols, obj.errors[0].keys)
  return obj, table






from test import run, nextexpid, run_params
from common import *
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.patches import Rectangle as Rect
from matplotlib import cm 


def run_synth(db, pp, ndim, volperc):
  params = []
  kdim = ndim
  npts = 2000
  for uo in (80,): # (30, 80)
    args = (ndim, kdim, npts, int(100.*volperc), uo)
    tablename = 'data_%d_%d_%d_0d%d_%duo' % args
    params.append((tablename, (npts, ndim, kdim, volperc, 10, 10, uo, 10)))

  for tablename, args in params:
    if ndim == 2 and kdim == 1:
        continue

    cs = [.61, .62, .63, .64, .65, .66, .67, .68, .69]
    expid = nextexpid(db)
    #run_params(db, tablename, expid, pp, klassname='Naive', klass=Naive, max_wait=45*60, cs=cs, granularity=10, naive=True, notes=tablename, p=.6, cols=None, l=.5, max_complexity=4)

    #run(db, tablename, pp, klasses=[Naive], max_wait=45*60, c=cs, cs=cs, granularity=15, naive=True, notes=tablename)
    #run(db, tablename, pp, klasses=[NDT],  c=cs, granularity=20, naive=True, notes=tablename)
    #run(db, tablename, pp, klasses= [BDT], l=[0.5], c=cs, epsilon=[0.0001, 0.001], tau=[0.1, 0.55], p=0.7, notes=tablename)
    #run(db, tablename, pp, klasses= [BDT], l=[0.5], c=[1.], epsilon=[0.01, 0.1], tau=[0.1, 0.55], p=0.7, notes=tablename)
    run(db, tablename, pp, klasses=[MR], l=[.5], c=cs, granularity=20, notes=tablename)

if __name__ == '__main__':
  from sqlalchemy import create_engine
  db = create_engine('postgresql://localhost/sigmod')
  pp = PdfPages('figs/test.pdf')

  print "did you run setup_experiments.py to generate the testing env?"

  run_synth(db, pp, 2, 0.5)
  #run_synth(db, pp, 3, 0.5)
  #run_synth(db, pp, 4, 0.5)

  pp.close()





#
# This file implements a naive decision tree algorithm that 
# 1) labels points in the bad tables by using a cutoff 
#    2 std from the mean tuple influence value
# 2) learns a decision tree using Orange
# 3) sorts the leaves by influence and returns the top k
#

import time
import pdb
import sys
import orngTree
import Orange
import orange
import heapq
import numpy as np
sys.path.extend(['.', '..'])

from itertools import chain
from sklearn import tree as sktree


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from basic import Basic
from merger import Merger

inf = 1e10000000
_logger = get_logger()

class NDT(Basic):

    def __init__(self, **kwargs):
        Basic.__init__(self, **kwargs)
        # values: c45, or, dt, rt
        self.tree_alg = kwargs.get('tree_alg', 'rt')

    def __call__(self, full_table, bad_tables, good_tables, **kwargs):
        """
        table has been trimmed of extraneous columns.
        """
        self.setup_tables(full_table, bad_tables, good_tables, **kwargs)

        self.SCORE_ID = add_meta_column(
                chain(self.bad_tables, self.good_tables),
                SCORE_VAR)
        self.CLASS_ID = add_meta_column(
                chain(self.bad_tables, self.good_tables),
                "INFCLASS",
                vals=['0', '1'])

        start = time.time()
        self.compute_perrow_influences(self.bad_tables, self.bad_err_funcs)
        self.compute_perrow_influences(self.good_tables, self.good_err_funcs)
        self.cost_compute_inf = time.time() - start



        start = time.time()
        if self.tree_alg == 'c45':
          table, rules = self.c45_rules()
        elif self.tree_alg == 'or':
          table, rules = self.orange_dt_rules()
        elif self.tree_alg == 'dt':
          table, rules = self.sk_dt_rules(max_depth=12)
        elif self.tree_alg == 'rt':
          table, rules = self.sk_rt_rules(max_depth=12)
        else:
          _logger.warn("unknown NDT algorithm %s.  Defaulting to regression tree", self.tree_alg)
          table, rules = self.sk_rt_rules(max_depth=12)
        self.cost_learn = time.time() - start


        #
        # ok now convert rules to clusters
        #

        _logger.debug( "got %d rules", len(rules))
        fill_in_rules(rules, table, cols=self.cols)

        self.cost_learn = time.time() - start

        clusters = [Cluster.from_rule(rule, self.cols) for rule in rules]
        for cluster in clusters:
          cluster.error = self.influence_cluster(cluster)
        clusters = filter_bad_clusters(clusters)
        clusters.sort(key=lambda c: c.error, reverse=True)
        print '\n'.join(map(str, clusters[:5]))

        self.all_clusters = self.final_clusters = clusters
        return self.final_clusters

        #
        # merge the clusters
        #
        thresh = compute_clusters_threshold(clusters, nstds=1.5)
        is_mergable = lambda c: c.error >= thresh

        params = dict(kwargs)
        params.update({
          'cols' : self.cols,
          'err_func' : self.err_func,
          'influence' : lambda c: self.influence_cluster(c),
          'influence_components': lambda c: self.influence_cluster_components(c),
          'is_mergable' : is_mergable,
          'use_mtuples' : False,
          'learner' : self})
        self.merger = Merger(**params)
        merged_clusters = self.merger(clusters)
        merged_clusters.sort(key=lambda c: c.error, reverse=True)


        clusters.extend(merged_clusters)
        normalize_cluster_errors(clusters)
        clusters = list(set(clusters))
        self.all_clusters = clusters
        self.final_clusters = merged_clusters

        self.costs = {
            'cost_learn' : self.cost_learn
        }
        return self.final_clusters

    def orange_dt_rules(self):

      start = time.time()
      bad_cutoff = self.influence_cutoff(self.bad_tables)
      good_cutoff = self.influence_cutoff(self.good_tables)
      _logger.debug( "cutoffs\t%f\t%f" , bad_cutoff, good_cutoff)
      self.cost_cutoff = time.time() - start

      _logger.debug( "creating training data")
      training = self.create_training(bad_cutoff, good_cutoff)


      #_logger.debug( "training on %d points" , len(training))
      tree = orngTree.TreeLearner(training)
      rules = tree_to_clauses(training, tree.tree)
      #_logger.debug('\n'.join(map(lambda r: '\t%s' % r, rules)))

      # tree = Orange.classification.tree.C45Learner(training, cf=0.001)
      # rules = c45_to_clauses(training, tree.tree)
      return training, rules



    def c45_rules(self):
      start = time.time()
      bad_cutoff = self.influence_cutoff(self.bad_tables)
      good_cutoff = self.influence_cutoff(self.good_tables)
      _logger.debug( "cutoffs\t%f\t%f" , bad_cutoff, good_cutoff)
      self.cost_cutoff = time.time() - start

      _logger.debug( "creating training data")
      training = self.create_training(bad_cutoff, good_cutoff)

      tree = Orange.classification.tree.C45Learner(training, cf=0.001)
      rules = c45_to_clauses(training, tree.tree)
      return training, rules



    def sk_dt_rules(self, **kwargs):
      start = time.time()
      _logger.debug( "computing cutoffs" )
      bad_cutoff = self.influence_cutoff(self.bad_tables)
      good_cutoff = self.influence_cutoff(self.good_tables)
      _logger.debug( "cutoffs\t%f\t%f" , bad_cutoff, good_cutoff)
      self.cost_cutoff = time.time() - start


      _logger.debug( "creating training data")
      training = self.create_training(bad_cutoff, good_cutoff)

      class_var = training.domain[self.CLASS_ID]
      skdata = rm_attr_from_domain(training, [class_var])
      Xs = np.array([ [v.value for v in row] for row in skdata ])
      Ys = np.array([int(row[class_var].value) for row in training])

      clf = sktree.DecisionTreeClassifier(criterion='entropy')
      clf.fit(Xs, Ys)
      rules = self.sktree_to_rules(skdata, clf.tree_, **kwargs)
      return skdata, rules


    def sk_rt_rules(self, **kwargs):
      orTable, Xs, Ys = self.create_regression_training()
      clf = sktree.DecisionTreeRegressor(
          criterion='mse',
          min_samples_split=10,
          min_samples_leaf=5
      )
      clf.fit(Xs, Ys)
      rules = self.sktree_to_rules(orTable, clf.tree_, **kwargs)
      return orTable, rules



    # assumes everything is continuous!!!
    def sktree_to_rules(self, table, tree, node_id=0, parent=None, clauses=None, max_depth=9, bdists=None):
      def clauses_to_rule(table, bdists, clauses):
        attr_to_ranges = {}
        for attridx, minval, maxval in clauses:
          attr = table.domain[attridx]
          dist = bdists[attr]
          cur = attr_to_ranges.get(attridx, [dist.min, dist.max])
          cur = (max(cur[0], minval), min(cur[1], maxval))
          attr_to_ranges[attridx] = cur

        conds = []
        for pos, (minv, maxv) in attr_to_ranges.iteritems():
          conds.append(Orange.data.filter.ValueFilterContinuous(
              position=pos,
              oper=orange.ValueFilter.Between,
              min=minv,
              max=maxv))
        
        return SDRule(table, None, conds)



      clauses = clauses or []
      ret = []
      left = tree.children_left[node_id]
      right = tree.children_right[node_id]

      if bdists is None:
        bdists = Orange.statistics.basic.Domain(table)
      

      if left < 0 or len(clauses) >= max_depth:
        val = tree.value[node_id]
        #if val[0][1] > val[0][0]:
        return [clauses_to_rule(table, bdists, clauses)]
      else:
        attridx = tree.feature[node_id]
        threshold = tree.threshold[node_id]
        leftclause = (attridx, -inf, threshold)
        rightclause = (attridx, threshold, inf)

        clauses.append(leftclause)
        ret.extend(self.sktree_to_rules(table, tree, left, node_id, clauses, max_depth, bdists))
        clauses.pop()

        clauses.append(rightclause)
        ret.extend(self.sktree_to_rules(table, tree, right, node_id, clauses, max_depth, bdists))
        clauses.pop()
      return ret


      
 
    def compute_perrow_influences(self, tables, err_funcs):
        infs = []
        for table, err_func in zip(tables, err_funcs):
          for row in table:
            influence = err_func([row])
            row[self.SCORE_ID] = inluencef
            infs.append(influence)
        return infs

    def influence_cutoff(self, tables, percentile=80):
        infs = []
        for table in tables:
          for row in table:
            infs.append(float(row[self.SCORE_ID]))

        #u = np.mean(infs)
        std = np.std(infs)
        #return u + std * 2
        return np.percentile(infs, percentile)

    def label_bad_tuples(self, cutoff):
        pass


    def create_regression_training(self):
        domain = Orange.data.Domain(self.bad_tables[0].domain)
        score_var = domain[self.SCORE_ID]
        newtable = Orange.data.Table(domain)

        Ys = []
        for t in chain(self.bad_tables, self.good_tables):
          newtable.extend(t)
          for r in t:
            Ys.append(float(r[self.SCORE_ID].value))

        Xs = np.array([ [v.value for v in row] for row in newtable])
        Ys = np.array(Ys)
        return newtable, Xs, Ys


    def create_training(self, bad_cutoff, good_cutoff):
        import orngTree
        extend_bad = lambda rule, t: rule.cloneAndAddContCondition(
                t.domain[self.SCORE_ID],
                bad_cutoff,
                1e100000)
        extend_good = lambda rule, t: rule.cloneAndAddContCondition(
                t.domain[self.SCORE_ID],
                -1e100000,
                good_cutoff)

        domain = self.bad_tables[0].domain
        score_var = domain[self.SCORE_ID]
        class_var = domain[self.CLASS_ID]
        domain = list(self.bad_tables[0].domain)
        domain = [a for a in domain if a.name in self.cols]
        domain = Orange.data.Domain(domain, class_var)
        domain.add_meta(self.SCORE_ID, score_var)
        self.CLASS_ID = 'INFCLASS'

        train_table = Orange.data.Table(domain)

        for table in self.bad_tables:
            rule = SDRule(table, None)
            bad_rule = extend_bad(rule, table)
            pos_matches = Orange.data.Table(domain,bad_rule.filter_table(table))
            neg_matches =  Orange.data.Table(domain,bad_rule.cloneAndNegate().filter_table(table))

            for row in pos_matches:
                row[class_var] ='1'
            for row in neg_matches:
                row[class_var] ='0'

            train_table.extend(pos_matches)
            train_table.extend(neg_matches)
        return train_table

        for table in self.good_tables:
            rule = SDRule(table, None)
            good_rule = extend_good(rule, table)
            matches = Orange.data.Table(domain, good_rule.filter_table(table))
            for row in matches:
                row[class_var] = '0'
            train_table.extend(matches)

        return train_table



def c45_to_clauses(table, node, bdists=None, clauses=None):
    clauses = clauses or []
    if not node:
      return []
    if bdists is None:
      bdists = Orange.statistics.basic.Domain(table)

    
    if node.node_type == 0: # Leaf
        quality = node.class_dist[1] 
        if int(node.leaf) == 1 and node.items > 0 and clauses is not None:
            ret = [rule_from_clauses(table, clauses)]
            for rule in ret:
                rule.quality = quality
            return ret 
        return []

    var = node.tested
    ret = []


    if node.node_type == 1: # Branch
        for branch, val in zip(node.branch, var.values):
            clause = create_clause(table, var,  val, bdists)
            clauses.append( clause )
            ret.extend( c45_to_clauses(table, branch, bdists, clauses) )
            clauses.pop()

    elif node.node_type == 2: # Cut
        for branch, comp in zip(node.branch, ['<=', '>', '<', '>=']):
            clause = create_clause(table, var,  node.cut, bdists, comp)
            clauses.append( clause )
            ret.extend( c45_to_clauses(table, branch, bdists, clauses) )
            clauses.pop()

    elif node.node_type == 3: # Subset
        for i, branch in enumerate(node.branch):
            inset = filter(lambda a:a[1]==i, enumerate(node.mapping))
            inset = [var.values[j[0]] for j in inset]
            if len(inset) == 1:
                clause = create_clause(table, var, inset[0], bdists)
            else:
                clause = create_clause(table, var, inset, bdists)
            clause.append( clause )
            ret.extend( c45_to_clauses(table, branch, bdists, clauses) )
            clauses.pop()

    ret = filter(lambda c: c, ret)
    return ret



def create_clause(table, attr, val, bdists, cmp='='):
    cmps = ['<', '<=', '>', '>=', '=']
    if attr.varType == Orange.feature.Type.Discrete:
        if not isinstance(val, (list, tuple)):
            val = [val]
        vals = [orange.Value(attr, v) for v in val]
        filt = orange.ValueFilter_discrete(
            position = table.domain.index(attr),
            values = vals)
        return filt
    else:
        # it may be a discretized continuous condition (e.g., "<= 5")
        isnumerical = False
        for c in cmps:
            try:
                if val.startswith(c):
                    val = float(val.split(c)[1])
                    cmp = c
                    isnumerical = True
                    break
            except:
                pass

        if not isnumerical:
            val = float(val)

        bdist = bdists[attr]

        minv, maxv = bdist.min, bdist.max
        op = None
        if cmp == '>=':
            minv = val
        elif cmp == '>':
            minv = val
        elif cmp == '<=':
            maxv = val
        elif cmp == '<':
            maxv = val
        elif cmp == '=':
            maxv = minv = val
        else:
            raise

        return Orange.data.filter.ValueFilterContinuous(
            position=table.domain.index(attr),
            oper=orange.ValueFilter.Between,
            min=minv,
            max=maxv)


def rule_from_clauses(table, clauses):
    domain = table.domain
    pos_to_cont_cond = {}
    pos_to_disc_cond = {}
    for c in clauses:
        pos = c.position
        attr = domain[pos]
        if attr.varType == Orange.feature.Type.Discrete:
            if pos in pos_to_disc_cond:
                vals = pos_to_disc_cond[pos]
                vals = vals.intersection(set(map(int, c.values)))
            else:
                vals = set(map(int, c.values))
            pos_to_disc_cond[pos] = vals
        else:
            if pos in pos_to_cont_cond:
                minv, maxv = pos_to_cont_cond[pos]
                minv = max(minv, c.min)
                maxv = min(maxv, c.max)
            else:
                minv, maxv = c.min, c.max
            pos_to_cont_cond[pos] = (minv, maxv)

    conds = []
    for pos, (minv, maxv) in pos_to_cont_cond.iteritems():
        conds.append(Orange.data.filter.ValueFilterContinuous(
            position=pos,
            oper=orange.ValueFilter.Between,
            min=minv,
            max=maxv))
    for pos, vals in pos_to_disc_cond.iteritems():
        conds.append(Orange.data.filter.ValueFilterDiscrete(
            position=pos,
            values=[Orange.data.Value(domain[pos], v) for v in vals]))
    
    return SDRule(table, None, conds)


def tree_to_clauses(table, node, bdists=None, clauses=None, strclauses=None):
    clauses = clauses or []
    strclauses = strclauses or []
    if not node:
        return []
    if bdists is None:
      bdists = Orange.statistics.basic.Domain(table)

    ret = []
    if node.branch_selector:
        varname = node.branch_selector.class_var.name
        var = table.domain[varname]
        for branch, bdesc in zip(node.branches,
                                 node.branch_descriptions):
            if ( bdesc.startswith('>') or 
                 bdesc.startswith('<') or 
                 bdesc.startswith('=') ):
                clauses.append( create_clause(table, var, bdesc, bdists))
            else:
                clauses.append( create_clause(table, var, bdesc, bdists) )
            strclauses.append((varname, bdesc))
            ret.extend( tree_to_clauses(table, branch, bdists, clauses, strclauses) )
            clauses.pop()
    else:
        major_class = node.node_classifier.default_value
        #if major_class == '1' and clauses:
        ret.append(rule_from_clauses(table, clauses))

    ret = filter(bool, ret)
    return ret





import pdb
import random
import numpy as np
from sklearn.neighbors import *
from scorpion.arch import create_orange_table
from scorpion.sigmod import AdjacencyGraph
from scorpion.bottomup import Cluster
random.seed(0)

vals = map(str, range(20))


rows = [ [v] for v in vals ]
table = create_orange_table(rows, ['moteid'])
domain = table.domain

clusters = []
for i in xrange(100):
  randvals = map(int, random.sample(vals, random.randint(1, 3)))
  c = Cluster([], 0, [], discretes={'moteid':randvals})
  clusters.append(c)
graph = AdjacencyGraph(clusters, domain)

metric = DistanceMetric.get_metric('hamming')
''.join(map(str,graph.feature_mapper(clusters[0], 'moteid')))
for n in graph.neighbors(clusters[0]):
  dist = metric.pairwise([
    graph.feature_mapper(clusters[0], 'moteid'),
    graph.feature_mapper(n, 'moteid')
  ])
  print '\t', dist[0], '\t', ''.join(map(str,graph.feature_mapper(n, 'moteid')))




import orange
import Orange
import math
from itertools import chain


def discrete_columns(table, cols=None):
    if cols is None:
        cols = [attr.name for attr in table.domain]

    ret = []
    for col in cols:
        attr = table.domain[col]
        if attr.varType == orange.VarTypes.Discrete:
            ret.append(col)
    return ret


def continuous_columns(table, cols=None):
    if cols is None:
        cols = [attr.name for attr in table.domain]

    ret = []
    for col in cols:
        attr = table.domain[col]
        if attr.varType != orange.VarTypes.Discrete:
            ret.append(col)
    return ret

        
def column_positions(table, cols):
    return [table.domain.index(table.domain[col]) for col in cols]


def valid_table_cols(table, cols, kwargs={}):
  bad_attrs = [
    'id', 'err', 'pickup_id', 'pickup_address', 'epoch', 'userid', 
    'mid', 'imdb', 'tstamp', "unknown", "action", "adventure", 
    "animation", "children", "comedy", "crime", "documentary", 
    "drama", "fantasy", "noir", "horro", "musical", "mystery", 
    "romance", "scifi", "thriller", "war", "western", 'lin_ima', 'com_nam',
    'total_charges', 'total_cost',  "totalcosts_ia", "totalcharges_ia",
    "estimatednetrevenue_ia", "estimated_net_revenue", "lengthofstay"
  ]
  ignore_attrs = kwargs.get('ignore_attrs', [])
  print "ignoring attrs: %s" % str(ignore_attrs)
  attrs = table.domain
  ret = []
  for attr in attrs:
    if attr.name in bad_attrs:
      continue
    if attr.name in cols:
      continue
    if attr.name in ignore_attrs:
      continue
    if attr.name.endswith('id') and attr.name != 'moteid':
      continue

    nunique = len(set([row[attr].value for row in table]))
    print "nunique %s\t%s" % (nunique, attr.name)

    if attr.varType != orange.VarTypes.Continuous:
      if nunique > 100 and nunique > 0.7 * len(table) or nunique > 7000:
        print "%s skipped" % attr.name
        continue
    ret.append(attr.name)
  return ret
 

def rm_attr_from_domain(table, attrs):
    """
    Remove attribute(s)/cols(s) from table
    Creates a copy of the original table :(
    """
    domain = table.domain
    try:
        erridxs = [domain.index(domain[attr]) for attr in attrs]
    except:
        erridxs = [domain.index(domain[attrs])]
    erridxs.sort(reverse=True)

    attrs = list(table.domain)
    map(attrs.pop, erridxs)
    
    newd = Orange.data.Domain(attrs)
    newd.add_metas(table.domain.get_metas())
    return Orange.data.Table(newd, table)
    # alternative way to create new table
    # return table.select_ref(newd)



def ids_filter(table, ids, attrname='id', negate=False):
    
    attr = table.domain[attrname]
    id_pos = table.domain.index(attr)
    vals = ids
    if not isinstance(ids[0], Orange.data.Value):
        vals = [Orange.data.Value(attr, id) for id in ids]
    if attr.var_type == Orange.feature.Type.Discrete:
        f = [ValueFilterDiscrete(position=id_pos, values=vals)]
    else:
        f = [ValueFilterContinuous(position=id_pos, ref=val, oper=ValueFilter.Equal) for val in vals]

    c = Orange.data.filter.Values(domain=table.domain,
                                  negate=negate,
                                  conjunction=0,
                                  conditions=f)
    def tmp(t, negate=negate):
        c.negate = negate
        return c(t)
    return tmp




def reconcile_tables(*all_tables):
  """
  Ensure bad and good tables use the same domain
  @return list of reconciled tables, union of all tables
  """
  domain = None
  full_table = None

  for table in chain(*all_tables):
    if domain is None:
      domain = list(table.domain.clone())
    else:
      for idx, (newattr, oldattr) in enumerate(zip(domain, table.domain)):
        if newattr.varType == orange.VarTypes.Discrete:
          if oldattr.varType == orange.VarTypes.Discrete:
            map(newattr.add_value, oldattr.values)
          else:
            domain[idx] = table.domain.clone()[idx]

  domain = Orange.data.Domain(domain)
  translated_tables = []

  for tables in all_tables:
    group = []
    translated_tables.append(group)
    
    for table in tables:
      #if not len(table):
      #continue
      
      if len(table):
        table = Orange.data.Table(domain)
      else:
        rows = [[v.value for v in row] for row in table]
        table = Orange.data.Table(domain, rows)

      group.append(table)
  
      if full_table is None:
        full_table = Orange.data.Table(domain, table)
      else:
        full_table.extend(table)
        

  if not full_table:
    return (), None

  return tuple(translated_tables), full_table

  translate = lambda t: Orange.data.Table(domain, t.to_numpyMA('ac')[0].data)
  translated_tables = []
  for tables in all_tables:#shared_domain_tables:
      translated_tables.append(map(translate, tables))

  return tuple(translated_tables), full_table


def union_tables(*tables):
  ret = None
  for t in chain(*tables):
    if ret is None:
      domain = t.domain.clone()
      ret = Orange.data.Table(domain)
    ret.extend(t)
  return ret

def add_meta_column(tables, name, vals=None, default=None):
    """
    @param name name of new attribute
    @param vals None if meta should be continuous, otherwise vals is a list of string values
           set to values of discrete attribute
    @param default default value.  
           if None, set to -inf for continuous attribute, or first element of vals for discrete
    """
    meta_id = Orange.feature.Descriptor.new_meta_id()
    for table in tables:
        if not vals:
            table.domain.addmeta(meta_id, Orange.feature.Continuous(name))
            table.add_meta_attribute(meta_id, default or -1e1000000)
        else:
            table.domain.addmeta(meta_id, Orange.feature.Discrete(name, values=vals))
            table.add_meta_attribute(meta_id, default or vals[0])
    return meta_id





"""
Messatfa - Comparative Study of Clustering Methods
multidimensional data generation

Basic Steps
1) Generate attribute definitions 
   - # attributes
   - name
   - type
   - value range
   - can default to [0,1] for everything
2) Define c clusters
   - split each attribute into c nonoverlapping attributes
   - each cluster doesn't overlap any range
3) Generate
   - same number of tuples per cluster

Extensions
- support values, not just tuple density

"""
import sys
import random
import numpy as np

random.seed(0)
np.random.seed(0)

class Attr(object):
    def __init__(self, name, vals=(0.,100.)):
        self.name = name
        self.vals = vals

    def ranges(self, c):
        mv = self.vals[0]
        vrange = self.vals[1] - self.vals[0]
        rsize = vrange / c
        return [Attr(self.name, vals=(mv+rsize*i, mv+rsize*(i+1)))
                for i in xrange(c)]
        
    def random(self, **kwargs):
        return random.uniform(*self.vals)

    def __str__(self):
        if self.vals[0] != 0 or self.vals[1] != 100: 
            return '(%3d-%3d)' % (self.vals[0], self.vals[1])
        return '(   -   )'


class ValAttr(Attr):

    def ranges(self, c):
        return [ValAttr(self.name, vals=self.vals) for i in xrange(c)]

    def random(self, bad=False):
        mv = self.vals[0]
        vrange = self.vals[1] - mv
        if bad:
            return random.uniform(vrange*0.4+mv, vrange*0.5+mv)#self.vals[1])
        return random.uniform(mv, vrange*0.2+mv)


           
class CDef(object):
    def __init__(self, attrs):
        self.attrs = attrs

    def random(self, **kwargs):
        return tuple(attr.random(**kwargs) for attr in self.attrs)


def gen_attributes(nattrs):
    """
    @param nattrs number of attribute
    """
    attrs = []
    for i in xrange(nattrs):
        attr = 'a_%d' % i
        attrs.append(Attr(attr, vals=(0, 100.)))
    attrs.append(ValAttr('val', vals=(0., 100.)))
    return attrs

def gen_cluster_defs(attrs, c):
    attr_ranges = []
    for attr in attrs:
        attr_ranges.append(attr.ranges(c))

    map(random.shuffle, attr_ranges)
    clusters = []
    for i in xrange(c):
        cdef = CDef(tuple(ar[i] for ar in attr_ranges))
        clusters.append(cdef)

    return clusters

def gen_dirty_cdefs(attrs, c):
    attr_ranges = []
    for attr in attrs:
        attr_ranges.append(attr.ranges(c))

    clusters = []
    for i in xrange(c):
        nattrs = random.randint(1, len(attrs))
        tmp = range(len(attrs))
        random.shuffle(tmp)
        tmp = sorted(tmp[:nattrs])

        cdef_attrs = []
        for idx in xrange(len(attrs)):
            if idx in tmp:
                cdef_attrs.append(random.choice(attr_ranges[idx]))
            else:
                cdef_attrs.append(attrs[idx])
        cdef = CDef(cdef_attrs)
        clusters.append(cdef)

    return clusters

       

def gen_data(cdefs, n, **kwargs):
    for cdef in cdefs[:len(cdefs)-1]:
       for i in xrange(n):
            yield cdef.random()
    cdef = cdefs[-1]
    for i in xrange(n):
        yield cdef.random(**kwargs)

def print_cdefs(cdefs):
    for idx, cdef in enumerate(cdefs):
        params = (idx == len(cdefs)-1 and 'Bad' or 'Good', 
                  '\t'.join(map(str, cdef.attrs)))
        print >>sys.stderr, '%s\t%s' % params

def print_data(idx, pts):
    for pt in pts:
        print '%d\t%s' % (idx, '\t'.join(map(lambda v: '%f' % v, pt)))


if __name__ == '__main__':
    if len(sys.argv) < 4:
        print "Generates and prints points in tab delimitied format"
        print "Usage:"
        print "\tpython genmultdimdata.py nattrs nclusters pts/cluster [niters]"
        exit()

    nattrs = int(sys.argv[1])
    nclusters = int(sys.argv[2])
    npts = int(sys.argv[3])
    niter = len(sys.argv) > 4 and int(sys.argv[4]) or 2


    # want to generate good and bad iterations using the same
    # clusters.
    # We should pick good and bad clusters and 
    # generate good and bad values for those clusters

    attrs = gen_attributes(nattrs)
    cdefs = gen_cluster_defs(attrs, nclusters)
    cdefs = gen_dirty_cdefs(attrs, nclusters)

    print_cdefs(cdefs)
    
    print 'iter\t%s' % '\t'.join(map(lambda a: a.name, attrs))
    for idx in xrange(niter):
        print_data(idx, gen_data(cdefs, npts, bad=(idx < niter/2)))



#
# This file used to be a parallel implementation of scorpion, 
# now it is the bare bones necessary code to run scorpion 
# single threaded
#

import random
import sys
import time
import pdb
import traceback
import numpy as np

from sklearn.cluster import KMeans
from itertools import chain
from collections import defaultdict
from datetime import datetime
from multiprocessing import Process, Queue, Pool, Pipe
from Queue import Empty

from scorpionsql.db import *
from scorpionsql.aggerror import *
import scorpionsql.errfunc as errfunc

from arch import *
from util import *
from sigmod import *
from settings import ID_VAR
from sigmod.streamrangemerger import *


_logger = get_logger()

DEFAULT_PARAMS = {
  'c_range': [0.05, 1],
  'l' : 0.6,
  'msethreshold': 0.01,
  'max_wait':5,
  'granularity': 50,
  'use_mtuples': False,#True,
  'DEBUG': False,

  # the following are for BDT
  'epsilon': 0.0015,
  'min_improvement': 0.01,
  'tau': [0.08, 0.5],
  'p': 0.7
}

def pick(l, key):
  return [v[key] for v in l]

def parallel_debug(sharedobj, **kwargs):
  for aggerr in sharedobj.errors:
    runner(sharedobj, aggerr, **kwargs)
  

def runner(sharedobj, aggerr, **kwargs):
  random.seed(2)
  sql = sharedobj.sql
  badresults = aggerr.keys
  label = aggerr.agg.shortname
  if not badresults:
      return

  try:
    table, rules, top_k_rules = serial_hybrid(sharedobj, aggerr, **kwargs)
  except:
    traceback.print_exc()
    rules = []
    table = None

  sharedobj.merged_tables[label] = table
  sharedobj.rules[label] = rules
  sharedobj.top_k_rules[label] = top_k_rules
  return table, rules



def load_tables(obj, aggerr, **kwargs):
  db = connect(obj.dbname)
  obj.db = db
  obj.update_status('loading inputs into memory')

  # get the input records!
  start = time.time()
  all_keys = list(chain(aggerr.keys, obj.goodkeys[aggerr.agg.shortname]))
  all_tables = get_provenance_split(obj, aggerr.agg.cols, all_keys)
  bad_tables = all_tables[:len(aggerr.keys)]
  good_tables = all_tables[len(aggerr.keys):]

  _logger.debug("bad table counts:  %s" % ', '.join(map(str, map(len, bad_tables))))
  _logger.debug("good table counts: %s" % ', '.join(map(str, map(len, good_tables))))
  print "agg error %s \t %s" % (aggerr.agg, aggerr.errtype)

  rules = []


  obj.update_status('preparing inputs')
  start = time.time()

  # "id" column is special so we need to deal with it specially
  cols = valid_table_cols(bad_tables[0], aggerr.agg.cols, kwargs)
  all_cols = cols + aggerr.agg.cols        
  if 'id' not in all_cols:
    all_cols.append('id')
  torm = [attr.name for attr in bad_tables[0].domain if attr.name not in all_cols]

  bad_tables = [rm_attr_from_domain(t, torm) for t in bad_tables]
  good_tables = [rm_attr_from_domain(t, torm) for t in good_tables]
  all_full_table = union_tables(bad_tables, good_tables)
  full_table = union_tables(bad_tables)

  print "total # rows:\t%d" % len(full_table)
  _logger.debug("valid cols: %s" % cols)


  # make sure aggerr keys and tables are consistent one last time
  if len(bad_tables) != len(aggerr.keys):
    pdb.set_trace()
    raise RuntimeError("#badtables (%d) != #aggerr keys (%d)" % (len(bad_tables), len(aggerr.keys)))

  return cols, bad_tables, good_tables, full_table, all_full_table


def rules_to_clusters(rules, learner):
  clusters = []
  fill_in_rules(
    rules, 
    learner.full_table, 
    cols=learner.cols, 
    cont_dists=learner.cont_dists
  )

  for r in rules:
    c = Cluster.from_rule(r, learner.cols)

    if not c.inf_state:
      learner.influence_cluster(c)
    if not r.c_range or abs(r.c_range[0]) == float('inf'):
      c.c_range = list(learner.c_range)
    c.inf_func = learner.create_inf_func(c)
    clusters.append(c)
  return clusters




def serial_hybrid(obj, aggerr, **kwargs):
  costs = {}
  cols, bad_tables, good_tables, full_table, all_full_table = load_tables(obj, aggerr, **kwargs)

  params = dict(DEFAULT_PARAMS)
  params.update({
    'obj': obj,
    'aggerr':aggerr,
    'cols':cols,
    'c': obj.c
  })
  params.update(dict(kwargs))

  if aggerr.agg.func.__class__ in (errfunc.SumErrFunc, errfunc.CountErrFunc):
    klass = MR 
  else:
    klass = BDT
  if False:
    klass = SVM

  learner = klass(**params)
  learner.setup_tables(all_full_table, bad_tables, good_tables)
  simplify = lambda r: r.simplify(cdists=learner.cont_dists, ddists=learner.disc_dists) 
  parallel = params.get('parallel', False)
  print "executing in parallel?", parallel

  start = time.time()

  if parallel:
    clusters = []
    par2chq = Queue()
    ch2parq = Queue()
    args = (learner, aggerr, params, _logger, (par2chq, ch2parq))
    proc = Process(target=merger_process_f, args=args)
    proc.start()

    # loop
    all_json_pairs = []
    for pairs in learner(all_full_table, bad_tables, good_tables):
      if not pairs: continue

      json_pairs = [(r.to_json(), idxkey) for r, idxkey in pairs]
      _logger.debug("main\tsend to child proc %d rules" % len(json_pairs))
      try:
				par2chq.put(json_pairs, False)
				_logger.debug("main\tsent")
      except Queue.Full:
				all_json_pairs.extend(json_pairs)
				_logger.debug("main\itq full")

    _logger.debug("main\tsend last batch to merger %d rules", len(all_json_pairs))
    par2chq.put(all_json_pairs)
    _logger.debug("main\tsend to child proc DONE")
    par2chq.put('done')
    par2chq.close()
    learner.update_status("waiting for merging step to finish")

    # wait for results from merger to come back
    json_rules = ch2parq.get()
    rules = [SDRule.from_json(d, learner.full_table) for d in json_rules]

    top_k_json_rules = ch2parq.get()
    top_k_json_rules = { 
      key: pick(g, 1) 
      for key, g in groupby(top_k_json_rules, key=lambda p: p[0])
    }

    ch2parq.close()
    proc.join()

    clusters = instrument(rules_to_clusters)(learner, rules, learner)

  else:
    mparams = dict(params)
    mparams.update({
      'learner_hash': hash(learner),
      'learner' : learner,
      'partitions_complete': False
    })
    merger = PartitionedStreamRangeMerger(**mparams)

    allpairs = []
    for pairs in learner(all_full_table, bad_tables, good_tables):
      allpairs.extend(pairs)

    rules, keyidxs = zip(*allpairs)
    clusters = instrument(rules_to_clusters)(learner, rules, learner)

    pairs = zip(clusters, keyidxs)
    for key, g in groupby(pairs, key=lambda p: p[1]):
      clusters = pick(g, 0)
      merger.add_clusters(clusters, idx=0, partitionkey=key)


    while merger.has_next_task():
			if merger():
				learner.update_rules(
					aggerr.agg.shortname, 
					group_clusters(merger.best_so_far(), learner)
				)
    clusters = merger.best_so_far(True)

    c_vals = CheapFrontier.compute_normalized_buckets(25, clusters)
    c_vals = c_vals * r_vol(learner.c_range) + learner.c_range[0]
    top_k_json_rules = defaultdict(list)
    for c_val in c_vals:
      for c in merger.best_at_c(c_val):
        rule = simplify(c.rule)
        rule.quality = c.inf_func(c_val)
        top_k_json_rules[c_val].append(rule.to_json())
    
    merger.close()

  costs['rules_get'] = time.time() - start


  top_k_rules = defaultdict(list)
  for c_val, json_rules in top_k_json_rules.iteritems():
    for d in json_rules:
      rule = SDRule.from_json(d, data=full_table)
      top_k_rules[c_val].append(rule)

  clusters = list(clusters)

  print "found clusters"
  for c in clusters:
    print "\t", str(c)


  obj.update_status('clustering results')
  start = time.time()
  rules = group_clusters(clusters, learner)
  costs['rules_cluster'] = time.time() - start
  learner.update_rules(aggerr.agg.shortname, rules)

  rules.sort(key=lambda r: r.c_range[0])
  rules = map(simplify, rules)

  print "grouped rules"
  for rule in rules: 
    print '\t', str(rule)

  print "=== Costs ==="
  for key, cost in costs.iteritems():
    print "%.5f\t%s" % (cost, key)
  for key, (cost, count) in learner.stats.iteritems():
    print "%.5f\t%s\t%s" % (cost,key, count)

  return full_table, rules, top_k_rules



def merger_process_f(learner, aggerr, params, _logger, (in_conn, out_conn)):
  valid_cluster_f = lambda c: True
  status = Status(learner.obj.status.reqid)
  simplify = lambda r: r.simplify(cdists=learner.cont_dists, ddists=learner.disc_dists) 

  def update_status(msg):
    start = time.time()
    merged = list(merger.best_so_far(True))
    rules = group_clusters(merged, learner)
    rules = map(simplify, rules)
    status.update_rules(aggerr.agg.shortname, rules)
    _logger.debug("merger\t%s %d rules\t%.4f sec", msg, len(rules), time.time()-start)

  params = dict(params)
  params.update({
    'learner_hash': hash(learner),
    'learner' : learner,
    'partitions_complete': False,
    'valid_cluster_f': valid_cluster_f 
  })
  merger = PartitionedStreamRangeMerger(**params)

  # setup a merger
  seen = set()
  json_pairs = []
  done = False
  while not done or merger.has_next_task():
    json_pairs = []
    if not done:
      try:
        json_pairs = in_conn.get(False)
      except Empty:
        json_pairs = []
        pass

    if json_pairs == 'done': 
      done = True
      json_pairs = []

    try:
      if len(json_pairs) > 0:
        pairs = [(SDRule.from_json(d, learner.full_table), keyidx) for d, keyidx in json_pairs]
        pairs = filter(lambda p: hash(p[0]) not in seen, pairs)
        seen.update([hash(p[0]) for p in pairs])
        rules, idxkeys = tuple(zip(*pairs))
        clusters = instrument(rules_to_clusters)(merger, rules, learner)
        pairs = zip(clusters, idxkeys)

        added = set()
        for key, g in groupby(pairs, key=lambda p: p[1]):
          toadd = pick(g, 0)
          _logger.debug("merger\tproc calls add_clusters\t%d clusters", len(toadd))
          added.update(merger.add_clusters(toadd, idx=0, partitionkey=key))

        if added:
          update_status("added")

      if merger.has_next_task():
        _logger.debug("merger\tprocess tasks\t%d tasks left" % merger.ntasks)
        if merger():
          update_status("updated")
        else:
          _logger.debug("merger\tno improvements\t%d tasks left", merger.ntasks)

    except Exception as e:
      print "problem in merger process"
      print e
      import traceback
      traceback.print_exc()

  _logger.debug("merger\texited loop")


  best = merger.best_so_far(True)
  best_json_rules = [
    c.rule.to_json() 
    for c in best
  ]

  c_vals = CheapFrontier.compute_normalized_buckets(25, best)
  c_vals = c_vals * r_vol(learner.c_range) + learner.c_range[0]
  top_k_json_rules = []
  for c_val in c_vals:
    for c in merger.best_at_c(c_val):
      rule = simplify(c.rule)
      rule.quality = c.inf_func(c_val)
      top_k_json_rules.append((c_val, rule.to_json()))

  merger.close()
  out_conn.put(best_json_rules)
  out_conn.put(top_k_json_rules)
  in_conn.close()
  out_conn.close()
  status.close()

  for key, (cost, count) in merger.stats.iteritems():
    print "%.5f\t%s\t%s" % (cost, key, count)
 
  print "child DONE"





def group_clusters(clusters, learner):
  """
  filter subsumed rules (clusters) and group them by
  those with similar influence on the user selected results

  Return
    list of rule objects sorted by c_range[0]
  """
  # find hierarchy relationships
  child2parent = get_hierarchies(clusters)
  validf = lambda c: valid_number(c.c_range[0]) and valid_number(c.c_range[1])

  ignore = set()
  non_children = set()
  for c in clusters:
    if c in ignore:
      continue

    if len(child2parent[c]):
      for p in child2parent[c]:
        if contributing(c, p) > 0.6:
          non_children.add(c)
          ignore.add(p)
      _logger.debug("groupclust\trm child cluster\t%s", c)
    elif not validf(c):
      _logger.debug("groupclust\tc_range invalid \t%s", c)
    else:
      non_children.add(c)

  groups = []
  for key, group in group_by_inf_state(non_children, learner).iteritems():
    subgroups = group_by_tuple_ids(group)
    subgroup = filter(bool, map(merge_clauses, subgroups.values()))
    groups.append(subgroup)

  rules = filter(bool, map(group_to_rule, groups))
  rules = sort_rules(rules, learner)
  return rules

def contributing(child, parent):
  ratios = []
  for c, p in zip(child.inf_state[0], parent.inf_state[0]):
    ratio = (1. + c) / (1. + p)
    ratios.append(ratio)
  return np.percentile(ratios, 20) 

def get_hierarchies(clusters):
  """
  Return a child -> parent relationship mapping
  """
  clusters = list(clusters)
  child2parent = defaultdict(list)
  for idx, c1 in enumerate(clusters):
    for c2 in clusters[idx+1:]:
      if c1.contains(c2):
        child2parent[c2].append(c1)
      elif c2.contains(c1):
        child2parent[c1].append(c2)
  return child2parent

def sort_rules(rules, learner):
  """
  if cluster doesn't affect outliers enough (in absolute terms),
  then throw it away
  """
  THRESHOLD = 0.01
  mean_val = np.median([ef.value for ef in learner.bad_err_funcs])
  threshold = THRESHOLD * mean_val

  def f(c):
    infs = None
    if c.inf_state:
      infs = filter(lambda v: abs(v) != float('inf'), c.inf_state[0])
    if not infs:
      return -1e10
    return max(infs) - threshold
  h = lambda c: r_vol(c.c_range)

  rules = sorted(rules, key=f, reverse=True)
  return rules

def merge_clauses(clusters):
  """
  assuming the clusters match the same input records, combine their clauses
  into a single cluster
  """
  if len(clusters) == 0: 
    _logger.debug("groupclust\tmerge_clauses\t%d clusters", len(clusters))
    return None
  if len(clusters) == 1: 
    clusters[0].rule.c_range = clusters[0].c_range
    return clusters[0]


  mainc = clusters[0]
  domain = mainc.rule.data.domain
  conds = {}
  for c in clusters:
    for d in c.rule.cond_dicts:
      col = d['col']
      if col not in conds:
        conds[col] = d
        continue

      if d['type'] == 'num':
        conds[col]['vals'] = r_intersect(conds[col]['vals'], d['vals'])
      else:
        conds[col]['vals'] = set(conds[col]['vals']).intersection(d['vals'])

  j = mainc.rule.to_json()
  j['clauses'] = conds.values()
  rule = SDRule.from_json(j, data=mainc.rule.data)
  c = Cluster.from_rule(rule, mainc.cols)
  c.inf_func = mainc.inf_func
  c.error = mainc.error
  return c


def group_to_rule(clusters):
  """
  pick a representative cluster from the arguments and 
  return a single one
  """
  _logger.debug("groupclust\tgroup->rule\t%d clusters", len(clusters))
  if len(clusters) == 0: return None
  for c in clusters:
    _logger.debug("groupclust\t\t%s", c)

  clusters = sorted(clusters, key=lambda c: r_vol(c.c_range), reverse=True)
  rule = clusters[0].rule
  rule.c_range = list(clusters[0].c_range)
  rule.cluster_rules = []
  for c in clusters[1:]:
    rule.cluster_rules.append(c.rule)

  return rule

def group_by_inf_state(clusters, learner):
  """
  super stupid check to group rules that have 
  identical influence states
  """
  errtype = learner.bad_err_funcs[0].errtype.errtype
  get_errval = lambda ef: ef.value
  if errtype == ErrTypes.EQUALTO:
    get_errval = lambda ef: ef.errtype(ef.value, ef.errtype.erreq)
  efs = np.array(map(get_errval, learner.bad_err_funcs))
  blocks = efs / 30.
  def trunc((idx,v)):
    if abs(v) == float('inf'):
      return 'inf'
    return int(float(v) / blocks[idx])

  def f(c):
    bad_infs =  tuple(map(trunc, enumerate(c.inf_state[0])))
    good_infs = tuple(c.inf_state[2])
    return (bad_infs, good_infs)
  return group_by_features(clusters, f)

def group_by_tuple_ids(clusters):
  """
  group the clusters by the set of input tuples they match
  """
  def f(cluster):
    ids = [int(row['id'].value) for row in cluster.rule.examples]
    ids.sort()
    return hash(tuple(ids))
  return group_by_features(clusters, f)

def group_by_features(clusters, f):
  groups = defaultdict(list)
  for c in clusters:
    groups[f(c)].append(c)

  return groups



import json
import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from multiprocessing import Process, Queue, Pool, Pipe
from Queue import Empty
from collections import deque
from itertools import chain
from rtree.index import Index as RTree
from rtree.index import Property as RProp
from scorpionsql.errfunc import ErrTypes

from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from basic import Basic
from sampler import Sampler
from merger import Merger
from rangemerger import RangeMerger, RangeMerger2
from streamrangemerger import *
from frontier import Frontier
from bdtpartitioner import *

inf = float('inf')
_logger = get_logger()








class BDT(Basic):

  def __init__(self, **kwargs):
    Basic.__init__(self, **kwargs)
    self.all_clusters = []
    self.cost_split = 0.
    self.cost_partition_bad = 0.
    self.cost_partition_good = 0.
    self.cache = None
    self.use_mtuples = kwargs.get('use_mtuples', False)
    self.max_wait = kwargs.get('max_wait', None)


  def __hash__(self):
    components = [
      self.__class__.__name__,
      str(self.aggerr.__class__.__name__),
      str(set(self.cols)),
      self.epsilon,
      self.tau,
      self.p,
      self.err_func.__class__.__name__,
      self.tablename,
      self.aggerr.keys,
      self.max_wait
    ]
    components = map(str, components)
    return hash('\n'.join(components))


  def setup_tables(self, full_table, bad_tables, good_tables, **kwargs):
    Basic.setup_tables(self, full_table, bad_tables, good_tables, **kwargs)


    # this is to cache a row's influence 
    self.SCORE_ID = add_meta_column(
      chain([full_table], bad_tables, good_tables),
      SCORE_VAR
    )

    domain = self.full_table.domain
    attrnames = [attr.name for attr in domain]
    self.cont_dists = dict(zip(attrnames, Orange.statistics.basic.Domain(self.full_table)))
    self.disc_dists = dict(zip(attrnames, Orange.statistics.distribution.Domain(self.full_table)))

    self.bad_states = [ef.state(t) for ef, t in zip(self.bad_err_funcs, self.bad_tables)]
    self.good_states = [ef.state(t) for ef, t in zip(self.good_err_funcs, self.good_tables)]


  def nodes_to_clusters(self, nodes, table):
    clusters = []
    rules = []
    for node in nodes:
      node.rule.quality = node.influence
      rule = node.rule.simplify(table, self.cont_dists, self.disc_dists)
      rules.append(rule)

    fill_in_rules(rules, table, cols=self.cols)
    for rule in rules:
      cluster = Cluster.from_rule(rule, self.cols)
      cluster.states = node.states
      cluster.cards = node.cards
      clusters.append(cluster)
    return clusters


  def pick_clusters(self, clusters, nonleaves):
    """
    """
    _logger.debug("compute initial cluster errors. %d clusters", len(clusters))
    start = time.time()
    for c in clusters:
      c.error = self.influence_cluster(c)
      c.c_range = list(self.c_range)
      c.inf_func = self.create_inf_func(c)
    self.stats['init_cluster_errors'] = [time.time()-start, 1]

    self.update_status("computing frontier")
    _logger.debug("compute initial frontier")
    frontier,_ = Frontier(self.c_range, 0.001)(clusters)

    ret = list(frontier)
    _logger.debug("get nonleaves containing frontier")
    for nonleaf in nonleaves:
      for c in frontier:
        if nonleaf.contains(c):
          nonleaf.error = self.influence_cluster(nonleaf)
          ret.append(nonleaf)
          break

    self.update_status("expanding frontier (%d rules)" % len(ret))
    _logger.debug("second merger pass")
    return ret


  def create_rtree(self, clusters):
    if not len(clusters[0].bbox[0]):
        class k(object):
            def intersection(self, foo):
                return xrange(len(clusters))
        return k()

    ndim = len(clusters[0].bbox[0]) + 1
    p = RProp()
    p.dimension = ndim
    p.dat_extension = 'data'
    p.idx_extension = 'index'

    rtree = RTree(properties=p)
    for idx, c in enumerate(clusters):
        rtree.insert(idx, c.bbox[0] + (0,) + c.bbox[1] + (1,))
    return rtree

  
  @instrument
  def intersect(self, bclusters, hclusters):
    errors = [c.error for c in bclusters]
    u, std = np.mean(errors), np.std(errors)
    u = min(max(errors), u + std)
    bqueue = deque(bclusters)
    low_influence = []
#        low_influence = [c for c in bclusters if c.error < u]
    bqueue = deque([c for c in bclusters if c.error >= u])

    hclusters = [c for c in hclusters if c.error >= u]
    if not hclusters:
      for c in bclusters:
        c.bad_states = c.states
        c.bad_cards = c.cards
      return bclusters
    hindex = self.create_rtree(hclusters)
    ret = []


    while bqueue:
      c = bqueue.popleft()

      idxs = hindex.intersection(c.bbox[0] + (0,) + c.bbox[1] + (1,))
      hcs = [hclusters[idx] for idx in idxs]
      hcs = filter(c.discretes_intersect, hcs)
      hcs = filter(c.discretes_contains, hcs)

      if not hcs:
        c.bad_inf = c.error
        c.good_inf = None
        c.bad_states = c.states
        c.bad_cards = c.cards
        ret.append(c)
        continue

      # first duplicate c into clusters that have the same discrete values as hcs
      # and those that are not
      # put the diff clusters back
      matched = False
      for hc in hcs:
        # are they exactly the same? then just skip
        split_clusters = c.split_on(hc)

        if not split_clusters: 
          # hc was false positive, skip
          continue

        matched = True

        intersects, excludes = split_clusters
        if len(intersects) == 1 and not excludes:
          c.good_inf = hc.error
          c.bad_inf = c.error
          c.good_states = hc.states
          c.bad_states = c.states
          c.bad_cards = c.cards
          c.good_cards = [math.ceil(n * c.volume / hc.volume) for n in c.cards]
          ret.append(c)
          continue
        else:
          for cluster in chain(intersects, excludes):
            cluster.good_inf, cluster.bad_inf, cluster.error = hc.error, c.error, c.error
            cluster.states = c.states
            new_vol = cluster.volume
            cluster.cards = [math.ceil(n * new_vol / c.volume) for n in c.cards]

          bqueue.extendleft(intersects)
          bqueue.extendleft(excludes)
          break

      if not matched:
        c.bad_inf = c.error
        c.good_inf = -inf
        c.bad_states = c.states
        c.bad_cards = c.cards
        ret.append(c)

    _logger.info( "intersection %d untouched, %d split" , len(low_influence), len(ret))
    ret.extend(low_influence)
    return ret


  @instrument
  def get_partitions(self, full_table, bad_tables, good_tables, **kwargs):
    #clusters, nonleaf_clusters = self.load_from_cache()
    #if clusters:
    #  yield clusters
    #  return
    #  #return clusters, nonleaf_clusters

    max_wait = self.params.get('max_wait', kwargs.get('max_wait', None))

    bad_params = dict(self.params)
    bad_params.update(kwargs)
    bad_params.update({
      'SCORE_ID': self.SCORE_ID,
      'err_funcs': self.bad_err_funcs,
      'max_wait': max_wait * 2
    })
    good_params = dict(self.params)
    good_params.update(kwargs)
    good_params.update({
      'SCORE_ID': self.SCORE_ID,
      'err_funcs': self.good_err_funcs,
      'max_wait': max_wait * 2
    })

    if not self.parallel:
      self.update_status("running partitioners in serial")
      partitioner = BDTTablesPartitioner(**bad_params)
      partitioner.setup_tables(bad_tables, full_table)
      partitioner.stats = self.stats
      gen = partitioner()
      pairs = [(n.rule, isleaf) for n, isleaf in gen]
      yield pairs
      bound = partitioner.get_inf_bound()

      partitioner = BDTTablesPartitioner(**good_params)
      partitioner.setup_tables(good_tables, full_table)
      partitioner.stats = self.stats
      partitioner.update_inf_bound(bound)
      gen = partitioner()
      pairs = [(n.rule, isleaf) for n, isleaf in gen]
      yield pairs
      return


    self.update_status("running partitioners in parallel")
    bad_p2cq = Queue()
    bad_c2pq = Queue()
    good_p2cq = Queue()
    good_c2pq = Queue()
    bad_args = (
      'bad', bad_params, self, 
      bad_tables, full_table, self.stats, 
      (bad_p2cq, bad_c2pq)
    )
    good_args = (
      'good', good_params, self,
      good_tables, full_table, self.stats,
      (good_p2cq, good_c2pq)
    )

    bad_proc = Process(target=partition_f, args=bad_args)
    good_proc = Process(target=partition_f, args=good_args)
    bad_proc.start()
    good_proc.start()

    bdone = gdone = False
    while not bdone or not gdone:
      bnodes = []
      gnodes = []
      bound = None
      if not bdone:
        try:
          data = bad_c2pq.get(False)
          if data == 'done':
            self.update_status("done partitioning outlier datasets")
            bdone = True
          else:
            (bnodes, bound) = data
        except Empty:
          bnodes = []

      if not gdone:
        try:
          data = good_c2pq.get(False)
          if data == 'done':
            self.update_status("done partitioning normal datasets")
            gdone = True
          else:
            gnodes = data[0]
        except Empty:
          gnodes = []

      
      if not gdone:
        if bound:
          good_p2cq.put(('bound', bound))
        leaves = [d for d, isleaf in bnodes if isleaf]
        if leaves:
          good_p2cq.put(('leaf', leaves))

      pairs = []
      if bnodes:
        pairs.extend(bnodes)
      if gnodes:
        pairs.extend(gnodes)

      if pairs:
        dicts = [p[0] for p in pairs]
        #_logger.debug("part\tgot %s\t%s" % (len(pairs), map(hash, map(str, dicts))))
        pairs = [(SDRule.from_json(d, full_table), isleaf) for d, isleaf in pairs]
        yield pairs

    self.update_status("done partitioning")
    bad_p2cq.close()
    good_c2pq.close()
    bad_c2pq.close()
    good_c2pq.close()
    bad_proc.join()
    good_proc.join()

    return


    ## NOTE: if good partitioner starts with tree from bad partitioner then
    ##       no need to intersect their results
    ##clusters = self.intersect(bclusters, hclusters)

    #self.cache_results(clusters, nonleaf_clusters)




  def __call__(self, full_table, bad_tables, good_tables, **kwargs):
      """
      table has been trimmed of extraneous columns.
      """
      self.setup_tables(full_table, bad_tables, good_tables, **kwargs)

      for pairs in self.get_partitions(full_table, bad_tables, good_tables, **kwargs):
        yield pairs
      return


  @instrument
  def load_from_cache(self):
    import bsddb as bsddb3
    self.cache = bsddb3.hashopen('./dbwipes.cache')
    try:
      myhash = str(hash(self))
      if myhash in self.cache and self.use_cache:
        self.update_status("loading partitions from cache")
        dicts, nonleaf_dicts, errors = json.loads(self.cache[myhash])
        clusters = map(Cluster.from_dict, dicts)
        nonleaf_clusters = map(Cluster.from_dict, nonleaf_dicts)
        for err, c in zip(errors, chain(clusters, nonleaf_clusters)):
          c.error = err
        return clusters, nonleaf_clusters
    except Exception as e:
      print e
      pass
    finally:
      self.cache.close()
    return None, None

  @instrument
  def cache_results(self, clusters, nonleaf_clusters):
    import bsddb as bsddb3
    # save the clusters in a dictionary
    if self.use_cache:
      myhash = str(hash(self))
      self.cache = bsddb3.hashopen('./dbwipes.cache')
      try:
        dicts = [c.to_dict() for c in clusters]
        nonleaf_dicts = [c.to_dict() for c in nonleaf_clusters]
        errors = [c.error for c in chain(clusters, nonleaf_clusters)]
        self.cache[myhash] = json.dumps((dicts, nonleaf_dicts, errors))
      except:
        pass
      finally:
        self.cache.close()






import numpy
import json
import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from collections import deque
from itertools import chain
from rtree.index import Index as RTree
from rtree.index import Property as RProp
from Orange.classification import svm
from sklearn import svm as sksvm
from scorpionsql.errfunc import ErrTypes

from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *
from ..settings import *

from basic import Basic
from sampler import Sampler
from merger import Merger
from bdtpartitioner import *

inf = 1e10000000
_logger = get_logger()





class SVM(Basic):

    def __init__(self, **kwargs):
        Basic.__init__(self, **kwargs)
        self.all_clusters = []
        self.cost_split = 0.
        self.cost_partition_bad = 0.
        self.cost_partition_good = 0.
        self.cache = None
        self.use_mtuples = kwargs.get('use_mtuples', False)




    def setup_tables(self, full_table, bad_tables, good_tables, **kwargs):
        Basic.setup_tables(self, full_table, bad_tables, good_tables, **kwargs)


        self.SCORE_ID = add_meta_column(
          chain([full_table], bad_tables, good_tables),
          SCORE_VAR
        )

        domain = self.full_table.domain
        attrnames = [attr.name for attr in domain]
        self.cont_dists = dict(zip(attrnames, Orange.statistics.basic.Domain(self.full_table)))
        self.disc_dists = dict(zip(attrnames, Orange.statistics.distribution.Domain(self.full_table)))

        self.bad_states = [ef.state(t) for ef, t in zip(self.bad_err_funcs, self.bad_tables)]
        self.good_states = [ef.state(t) for ef, t in zip(self.good_err_funcs, self.good_tables)]

        self.good_table = Orange.data.Table(domain, good_tables[0])
        for gt in good_tables[1:]:
          self.good_table.extend(gt)

        self.bad_table = Orange.data.Table(domain, bad_tables[0])
        for bt in bad_tables[1:]:
          self.bad_table.extend(bt)



    def learn(self, data, nu=0.01, kernel=svm.kernels.RBF, svmtype=svm.SVMLearner.OneClass):
        learner = svm.SVMLearner(svm_type=svmtype, kernel_type=kernel, nu=nu)
        classifier = learner(data)
        return classifier


    def __call__(self, full_table, bad_tables, good_tables, **kwargs):
        """
        table has been trimmed of extraneous columns.
        """
        self.setup_tables(full_table, bad_tables, good_tables, **kwargs)


        gooddata = numpy.array( [
          [float(v.value != 'None' and v.value or -1) for v in row] 
          for row in self.good_table
        ])

        allbaddata = numpy.array( [
          [float(v.value != 'None' and v.value or -1) for v in row] 
          for row in self.bad_table
        ])

        learner = sksvm.OneClassSVM(kernel='rbf', nu=0.01)
        sample = random.sample(gooddata, 8000)
        clf = learner.fit(sample)
        labels = clf.predict(allbaddata)
        bad = allbaddata[labels == -1]
        import pdb
        pdb.set_trace()

 
        classifier = self.learn(self.good_table)
        labels = map(int, map(classifier, self.bad_table))
        badrows = [self.bad_table[idx] for idx, label in enumerate(labels) if label == -1]
        sv = classifier.support_vectors
        datas = [ ['%4f' % float(v.value != 'None' and v.value or 0) for v in row] for row in sv]
        import pdb
        pdb.set_trace()

        # create a cluster
        return None



        clusters = self.get_partitions(full_table, bad_tables, good_tables, **kwargs)
        self.all_clusters = clusters

        start = time.time()
        _logger.debug('merging')
        self.final_clusters = self.merge(clusters)        
        filt = lambda c: not math.isinf(c.error) and not math.isnan(c.error)
        self.final_clusters = filter(filt, self.final_clusters)
        self.cost_merge = time.time() - start


        self.costs.update( {'cost_partition_bad' : self.cost_partition_bad,
                'cost_partition_good' : self.cost_partition_good,
                'cost_split' : self.cost_split,
                'cost_merge' : self.cost_merge})
        
        return self.final_clusters




from scorpion.db import connect
import random

db = connect('fec12')

for i in xrange(500):
  amt = 100000 + random.randint(-5000, 5000)
  nm = 'Obama, Barack'
  candid = 'P80003338'
  disb_dt = '2012-12-10'
  recipient_nm = "eugene"
  recipient_city = "Berkeley"
  recipient_st = "CA"
  recipient_zip = random.randint(94520, 94530)
  memo_text = "eugene"
  disb_desc = "eugene"
  cmte_id = "myid%d" % (i + 10000000)

  q = """INSERT INTO expenses_extradata(
     cmte_id, cand_id, cand_nm, recipient_nm,
     recipient_city, recipient_st, recipient_zip,
     memo_text, disb_desc, disb_dt, disb_amt) 
     values (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)"""
  db.execute(q, (
    cmte_id,
    candid,
    nm,
    recipient_nm,
    recipient_city,
    recipient_st,
    recipient_zip,
    memo_text,
    disb_desc,
    disb_dt,
    amt
  ))






import pdb
import sys
import random
import time
import sys
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.patches import Rectangle as Rect
from matplotlib import cm 


from scorpion.util.rangeutil import *

matplotlib.use("Agg")

class BaseRenderer(object):
  def __init__(self, fname, *args, **kwargs):
    self.pp = PdfPages(fname)
    self.clusters = []
    self.fig = plt.figure(figsize=(8, 8))
    self.sub = self.fig.add_subplot(111)
    self.xbound = None
    self.ybound = None

  def new_page(self):
    if (self.fig):
      self.fig.savefig(self.pp, format='pdf')
    self.sub.cla()

  def close(self):
    if self.fig:
      plt.savefig(self.pp, format='pdf')
    self.pp.close()
    self.pp = None
    self.fig = self.sub = None

  def set_lims(self, xbound=None, ybound=None):
    xbound = xbound or self.xbound
    ybound = ybound or self.ybound
    if xbound:
      xbound = r_expand(xbound)
      ybound = r_expand(ybound)
      self.sub.set_xlim(*xbound)
      self.sub.set_ylim(*ybound)


  def plot_title(self, title):
    self.sub.set_title(title)
    self.sub.title.set_fontsize(6)

  def set_title(self, title):
    self.sub.set_title(title)
    self.sub.title.set_fontsize(6)


class ClusterRenderer(BaseRenderer):
  def __init__(self, *args, **kwargs):
    BaseRenderer.__init__(self, *args, **kwargs)
    self.xbound = self.ybound = None

  def new_page(self):
    BaseRenderer.new_page(self)
  
  def transform_box(self, x, y):
    return x, y

  def plot_clusters(self, clusters, cols=None, color=None, alpha=None):
    self.clusters = clusters
    errors = [c.error for c in clusters]
    errors = np.array(errors)
    mean, std = np.mean(errors), np.std(errors)
    if std == 0: std = 1
    errors = (errors - mean) / std

    if not cols:
      cols = [0, 1]
    if isinstance(cols[0], basestring):
      cols = map(clusters.cols.index, cols)

    for idx, cluster in enumerate(clusters):
      tup = tuple(map(list, zip(*cluster.bbox)))
      x, y = tup[cols[0]], tup[cols[1]]
      x, y = self.transform_box(x, y)
      if not self.xbound:
        self.xbound = [x[0], x[1]]
        self.ybound = [y[0], y[1]]
      else:
        self.xbound = r_union(self.xbound, x)
        self.ybound = r_union(self.ybound, y)

      a = alpha or min(1, max(0.1, errors[idx]))
      c = color or cm.jet(errors[idx])
      r = Rect((x[0], y[0]), x[1]-x[0], y[1]-y[0], alpha=a, ec=c, fill=False, lw=1.5)
      self.sub.add_patch(r)

    self.set_lims()

  def plot_tuples(self, tuples):
    self.tuples = tuples
    cols = zip(*tuples)
    xs, ys, cs = cols[0], cols[1], cols[-1]
    self.sub.scatter(ys, xs, c=cs, alpha=0.5, lw=0)

    if not self.xbound:
      self.xbound = [min(xs), max(xs)]
      self.ybound = [min(ys), max(ys)]
    self.xbound = r_union(self.xbound, [min(xs), max(xs)])
    self.ybound = r_union(self.ybound, [min(ys), max(ys)])
    self.set_lims()


class JitteredClusterRenderer(ClusterRenderer):
  def transform_box(self, x, y):
    f = lambda v: (random.random()-0.5)*2+v
    x = [f(x[0]), f(x[1])]
    y = [f(y[0]), f(y[1])]
    return x, y


class InfRenderer(BaseRenderer):
  def __init__(self, *args, **kwargs):
    BaseRenderer.__init__(self, *args, **kwargs)
    self.c_range = kwargs.get('c_range', [0, 1])
    self.xbound = None
    self.ybound = None


  def new_page(self):
    BaseRenderer.new_page(self)

  def plot_points(self, xs, ys, color='grey', alpha=0.3):
    self.sub.plot(xs, ys, color=color, alpha=alpha)

    ys = filter(lambda v: abs(v) != float('inf'), ys)
    if len(ys): 
      if not self.ybound:
        self.xbound = [min(xs), max(xs)]
        self.ybound = [ min(ys), max(ys) ]
      else:
        inf_range = [min(ys), max(ys)]
        self.xbound = [min(self.xbound[0], min(xs)), max(self.xbound[1], max(xs))]
        self.ybound = [min(self.ybound[0], min(ys)), max(self.ybound[1], max(ys))]

    self.set_lims()



  def plot_active_inf_curves(self, clusters, color='red', alpha=0.3):
    alpha = alpha or 0.3
    color = color or 'red'
    for c in clusters:
      xs = (np.arange(100) / 100. * r_vol(c.c_range)) + c.c_range[0]
      ys = map(c.inf_func, xs)
      self.sub.plot(xs, ys, color=color, alpha=alpha)

      ys = filter(lambda v: abs(v) != float('inf'), ys)
      if not ys: continue
      if not self.ybound:
        self.xbound = c.c_range
        self.ybound = [ min(ys), max(ys) ]
      else:
        inf_range = [min(ys), max(ys)]
        self.xbound = [min(zip(self.xbound, c.c_range)[0]),  max(zip(self.xbound, c.c_range)[1])]
        self.ybound = [min(zip(self.ybound, inf_range)[0]),  max(zip(self.ybound, inf_range)[1])]

    self.set_lims()


  def plot_inf_curves(self, clusters, c_range=None, color=None, alpha=None):
    """
    By default, renders each cluster's inf curve and colors the valid 
    cluster.c_range segment as red.

    Optionally, can render all clusters the same color and alpha based on
    passed-in args
    """
    if not c_range: c_range = self.c_range
    xs = (np.arange(100) / 100. * r_vol(c_range)) + c_range[0]

    for c in clusters:
      alpha = alpha or 0.3
      color = color or 'grey'
      ys = map(c.inf_func, xs)
      self.sub.plot(xs, ys, color=color, alpha=alpha)

      ys = filter(lambda v: abs(v) != float('inf'), ys)
      if not ys: continue
      if not self.ybound:
        self.xbound = c_range
        self.ybound = [ min(ys), max(ys) ]
      else:
        inf_range = [min(ys), max(ys)]
        self.xbound = [min(zip(self.xbound, c_range)[0]),  max(zip(self.xbound, c_range)[1])]
        self.ybound = [min(zip(self.ybound, inf_range)[0]),  max(zip(self.ybound, inf_range)[1])]



    self.set_lims()



import sys
import random
import time
import sys
import matplotlib
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.patches import Rectangle as Rect
from matplotlib import cm 
from collections import defaultdict

from scorpionsql.db import *
from scorpionsql.aggerror import *
from scorpion.arch import *
from scorpion.tests.gentestdata import *
from scorpion.util import reconcile_tables
from scorpion.sigmod import *

matplotlib.use("Agg")

def get_row_ids(r, table):
    return set([row['id'].value for row in r(table)])

def get_tuples_in_bounds(db ,tablename, bounds, additional_where=''):
    with db.begin() as conn:
        where = ['%f <= a_%d and a_%d <= %f' % (minv, i, i, maxv) for i, (minv, maxv) in enumerate(map(tuple, bounds))]
        if additional_where:
            where.append(additional_where)
        where = ' and '.join(where)
        q = """ select * from %s where %s""" % (tablename, where)
        return [map(float, row) for row in conn.execute(q)]



def get_ids_in_bounds(db ,tablename, bounds):
  """
  Given a synthetic tablename and bounds of the form [ [min1, max1], ... ]
  Return the ids of the records that match
  """
  with db.begin() as conn:
      where = ['%f <= a_%d and a_%d <= %f' % (minv, i, i, maxv) for i, (minv, maxv) in enumerate(map(tuple, bounds))]
      where = ' and '.join(where)
      q = """ select id from %s where %s""" % (tablename, where)
      return [int(row[0]) for row in conn.execute(q)]




def compute_stats(found_ids, bad_tuple_ids, table_size):
    bad_tuple_ids = set(bad_tuple_ids)
    found_ids = set(found_ids)
    n = len(found_ids)        
    tp = len(bad_tuple_ids.intersection(found_ids))
    fn = len(bad_tuple_ids.difference(found_ids))
    fp = len(found_ids.difference(bad_tuple_ids))
    tn = table_size - tp - fn - fp

    accuracy = float(tp + tn) / table_size
    precision = float(tp) / (tp + fp) if n and (tp+fp) else 0.
    recall = float(tp) / (tp + fn) if n and (tp+fn) else 0.
    f1 = (precision*recall*2.) / (precision+recall) if precision+recall > 0 else 0.

    return accuracy, precision, recall, f1

def kname(klass):
    return klass.__name__[-7:]



def strip_columns(table, aggerr, cols=None):
    cols = cols or [attr.name for attr in table.domain]
    cols = [col for col in cols 
            if (col not in ['id', 'err', 'epochid', 'date', 'light'] and 
                col not in aggerr.agg.cols)]
    all_cols = cols + aggerr.agg.cols
    torm = [attr.name for attr in table.domain if attr.name not in all_cols]
    table = rm_attr_from_domain(table, torm)
    return table, cols



def get_parameters(datasetidx, **kwargs):
    name = datasetnames[datasetidx]
    outname = name

    test_data = get_test_data(name)
    dbname, sql, badresults, goodresults, errtype, get_ground_truth, tablename = test_data
    obj, table = create_sharedobj(*test_data[:-2])
    aggerr = obj.errors[0]

    # retrieve table for each good and bad key
    obj.db = connect(dbname)
    bad_tables = get_provenance_split(obj, aggerr.agg.cols, aggerr.keys) or []
    good_tables = get_provenance_split(obj, aggerr.agg.cols, obj.goodkeys[aggerr.agg.shortname]) or []

    (bad_tables, good_tables), full_table = reconcile_tables(bad_tables, good_tables)
    #_, full_table = reconcile_tables(bad_tables)

    # strip unnecessary columns
    user_cols = kwargs.get('cols', None)
    table, cols = strip_columns(table, aggerr, cols=user_cols)
    bad_tables = [strip_columns(t, aggerr, cols=user_cols)[0] for t in bad_tables]
    good_tables = [strip_columns(t, aggerr, cols=user_cols)[0] for t in good_tables]
    table = full_table

    truth = set(get_ground_truth(full_table))

    return full_table, bad_tables, good_tables, truth, aggerr, cols
 

def get_rules(full_table, bad_tables, good_tables, **kwargs):
    """
    Runs Scorpion on this dataset
    """
    klass = kwargs['klass']

    learner = klass(**kwargs)
    start = time.time()
    clusters = learner(full_table, bad_tables, good_tables, **kwargs)
    end = time.time()
    cost = end - start
    costs = learner.costs
    clusters = filter(bool, clusters)
    all_clusters = learner.all_clusters


    normalize = lambda arg: normalize_cluster_errors([c.clone() for c in arg])
    clusters, all_clusters = map(normalize, (clusters, all_clusters))
    best_clusters = filter_top_clusters(clusters, nstds=1)

    cols = kwargs['cols']
    
    thresh = compute_clusters_threshold(clusters)
    for c in clusters:
        c.isbest = (c.error >= thresh)
    merged = sorted(clusters_to_rules(clusters, full_table), key=lambda c: c.quality, reverse=True)
    #merged = [r.simplify() for r in merged]

    cost = cost - costs.get('merge_load_from_cache',(0,))[0] - costs.get('merge_cache_results', (0,))[0]
    costs['cost_merge'] = costs.get('cost_merge', 0) - (costs.get('merge_load_from_cache',(0,))[0] + costs.get('merge_cache_results', (0,))[0])
    costs['cost_total'] = cost
    return costs, merged, learner

def run_experiment(datasetidx, **kwargs):
    print kwargs
    ft,bts,gts, truth, aggerr, cols = get_parameters(datasetidx, **kwargs)

    params = {
        'epsilon':0.001,
        'tau':[0.01, 0.25],
        'lamb':0.5,
        'min_pts':3,
        'c' : 0.  
    }
    params.update(kwargs)
    params.update({
        'aggerr' : aggerr,
        'cols' : cols,
        'tablename' : datasetidx,
        'dataset' : datasetidx
        })

    costs, rules, learner = get_rules(ft, bts, gts, **params)
    ids = [get_row_ids(rule, ft) for rule in rules]

    learner.__dict__['compute_stats'] = lambda r: compute_stats(r, truth, len(ft))


    return costs, rules, ids, len(ft), learner


    f_stats = lambda r: compute_stats(r, truth, ft)

    bstats = map(f_stats, best)
    mstats = map(f_stats, merged)
    astats = map(f_stats, allr)
    return cost, costs, bstats, mstats, astats








from operator import mul, and_


def points_bounding_box(points):
    if not len(points):
        return ((0,), (0,))
    return (tuple(points.min(0)), tuple(points.max(0)))


def bounding_box(bbox1, bbox2):
    mins = tuple([min(min1, min2) for min1, min2 in zip(bbox1[0], bbox2[0])])
    maxs = tuple([max(max1, max2) for max1, max2 in zip(bbox1[1], bbox2[1])])
    return (mins, maxs)

def intersection_box(bbox1, bbox2):
    mins = map(max, zip(bbox1[0], bbox2[0]))
    maxs = map(min, zip(bbox1[1], bbox2[1]))
    return (mins, maxs)

def box_contained(box, bound, epsilon=0.):
    if not len(zip(*box)) and not len(zip(*bound)):
      return True
    if len(zip(*box)) != len(zip(*bound)):
      print >>sys.stderr, "box_contained: bounding boxes have different dimensions: %d vs %d" % (len(box[0]), len(bound[0]))
      return False
    inter = intersection_box(box, bound)
    idiffs = [p[1]-p[0] for p in zip(*inter)]
    bdiffs = [p[1]-p[0] for p in zip(*box)]
    return reduce(and_, map(lambda (i,b): i >= (1.-epsilon) * b, zip(idiffs, bdiffs)))

    return (reduce(and_, (min1 >= min2 for min1, min2 in zip(box[0], bound[0]))) and
            reduce(and_, (max1 <= max2 for max1, max2 in zip(box[1], bound[1]))) )

def box_completely_contained(box, bound, epsilon=0.):
    if not len(zip(*box)) or not len(zip(*bound)):
        return False
    for inner, outer in zip(zip(*box), zip(*bound)):
        if inner[0] <= outer[0] or inner[1] >= outer[1]:
            return False
        if (outer[1]-outer[0]) - (inner[1]-inner[0]) <= epsilon:
            return False
    return True


def box_same(box1, box2, epsilon=0.):
    ivol = volume(intersection_box(box1, box2))
    f = lambda box: ivol >= (1. - epsilon) * volume(box)
    return reduce(and_, map(f, (box1, box2)))

def volume(bbox):
    if not len(bbox[0]): return 0.
    diffs = [maxv - minv for minv, maxv in zip(*bbox)]
    if min(diffs) < 0: return 0
    return reduce(mul, diffs) 




from itertools import *

from ..learners.cn2sd.rule import SDRule
from ..util import powerset

INF = float('inf')

class Node(object):
    def __init__(self, rule):
        self.rule = rule
        self.children = []
        self.parent = None
        self.n = 0
        self.influence = -INF
        self.prev_attr = None
        self.score = -INF


        # boolean set if it was cloned from results of
        # partitioning outlier outputs.  See bdt.get_partitions()
        self.frombad = False


        self.cards = None # caches the cardinality of in each input group
        self.states = None # caches M-tuples

    def clone(self):
      """
      clones the structure of the tree, but loses score/influence info
      """
      node = Node(self.rule.clone())
      for child in self.children:
        child = child.clone()
        child.parent = node
        node.add_child(child)
      return node

    def set_score(self, score):
      if self.influence == -INF or self.influence > score:
        self.influence = score
        if self.parent:
          self.parent.set_score(score)

    def add_child(self, child):
        self.children.append(child)
    
    def __depth__(self):
        if not self.parent:
          return 0
        return 1 + self.parent.depth
    depth = property(__depth__)

    def __path__(self, q=None):
      q = q or list()
      if self.parent:
        self.parent.__path__(q)
      q.append(self)
      return q
    path = property(__path__)
        

    def __leaves__(self):
      if self.n is None:
        return []
      if not self.children:
        return [self]
      return chain(*[child.leaves for child in self.children])
    leaves = property(__leaves__)

    def __nonleaves__(self):
      """
      @deprecated
      Return ancestors of the leaf nodes found in this tree
      """
      if self.n is None:
        return []
      if not self.children:
        return []
      nodes = [child.nonleaves for child in self.children] + [[self]]
      return chain(*nodes)

    #def __nonleaves__(self):
    #  """
    #  Return all possible parents of the leaf nodes
    #  e.g., if leaf node has conditions (a,b,c)
    #        returns (a), (b), (c), (a,b), (a,c), (b,c), (a,b,c)
    #  """
    #  print "get nonleaves"
    #  ret = set()
    #  for leaf in self.leaves:
    #    conds = leaf.rule.filter.conditions
    #    for subset in powerset(conds):
    #      if len(subset) == len(conds): continue
    #      newrule = SDRule(leaf.rule.data, None, subset, leaf.rule.g)
    #      ret.add(newrule)
    #  print "done"
    #  return map(Node, ret)
    nonleaves = property(__nonleaves__)



    def __nodes__(self):
      if self.n is None:
        return []
      allnodes = [[self]] + [child.nodes for child in self.children]
      return chain(*allnodes)
    nodes = property(__nodes__)

    def __str__(self):
        return '%.4f\t%d\t%s' % (self.influence, self.n, self.rule)





#
# Generates random data for SYNTH dataset in paper
# N dimensions, 1 group-by dim (5 outlier, 5 normal groups), 1 value attr
# Normal points draw value from N(u_h, s_h) -- gaussian, u_h mean, s_h std
# Outlier points draw value from N(u_o, s_o)
# Outlier points fall in a 10% (variable) volume box in k<=N dimensions
#
import pdb
import sys
import random
from operator import and_, mul
random.seed(0)

def rand_box(ndim, kdim, vol, bounds=None):
    """
    outlier attributes are a_1,...a_kdim
    """
    if not bounds:
        bounds = [(0, 100)] * ndim
    bedges = [b[1]-b[0] for b in bounds]

    totalvol = reduce(mul, bedges)
    usedvol = reduce(mul, bedges[kdim:], 1)
    availvol = max(0, vol * totalvol / usedvol)
    edge = availvol ** (1. / kdim)

    ret = []
    for attr, bound in zip(xrange(ndim), bounds):
        if attr < kdim:
            lower = random.random() * (bound[1]-bound[0] - edge) + bound[0]
            upper = lower + edge
        else:
            lower, upper = bound[0], bound[1]
        ret.append((lower, upper))
    return ret 


def in_box(pt, box):
    return reduce(and_, [l <= v and v <= u for v, (l, u) in zip(pt, box)])

def rand_point(ndim):
    return [random.random() * 100 for i in xrange(ndim)]


def gen_multi_outliers(npts, ndim, kdim, vol, uh=10, sh=5, uo=90, so=5):
  # completely reproducable
  random.seed(0)

  nclusters = 2
  norm_gen, mid_gen, outlier_gen = make_gen(uh, sh), make_gen((uo-uh)/2+uh, so),  make_gen(uo, so)
  med_boxes = [rand_box(ndim, kdim, vol/2.) for i in xrange(nclusters)]
  high_boxes = [rand_box(ndim, kdim, vol, bounds=med_box) for med_box in med_boxes]

  for med_box in med_boxes:
    print >>sys.stderr, map(lambda arr: map(int, arr), med_box)
  for high_box in high_boxes:
    print >>sys.stderr, map(lambda arr: map(int, arr), high_box)


  schema = ['a_%d' % i for i in xrange(ndim)] + ['g', 'v']

  def generate():
    for gid in xrange(10):
      for i in xrange(npts):
        pt = rand_point(ndim)

        # add group and value
        pt.append(gid)

        if gid >= 5 and any([in_box(pt, mb) for mb in med_boxes]):
          if any([in_box(pt, hb) for hb in high_boxes]):
            pt.append(outlier_gen())
          else:
            pt.append(mid_gen())
        else:
          pt.append(norm_gen())

        yield pt

  return med_boxes, high_boxes, schema, generate()



def gen_points(npts, ndim, kdim, vol, uh=10, sh=5, uo=90, so=5):
  # completely reproducable
  random.seed(0)

  norm_gen, mid_gen, outlier_gen = make_gen(uh, sh), make_gen((uo-uh)/2+uh, so),  make_gen(uo, so)
  outlier_box = rand_box(ndim, kdim, vol)
  super_box = rand_box(ndim, kdim, vol, bounds=outlier_box)
  print >>sys.stderr, map(lambda arr: map(int, arr), outlier_box)
  print >>sys.stderr, map(lambda arr: map(int, arr), super_box)


  schema = ['a_%d' % i for i in xrange(ndim)] + ['g', 'v']

  def generate():
    for gid in xrange(10):
      for i in xrange(npts):
        pt = rand_point(ndim)

        # add group and value
        pt.append(gid)

        if gid >= 5 and in_box(pt, outlier_box):
          if in_box(pt, super_box):
              pt.append(outlier_gen())
          else:
              pt.append(mid_gen())
        else:
          pt.append(norm_gen())

        yield pt

  return outlier_box, super_box, schema, generate()


def make_gen(u, s):
    return lambda: random.gauss(u, s)



if __name__ == '__main__':
    import sys
    if len(sys.argv) < 5:
        print "Generates SYNTH dataset for paper. (single k dim sub-cluster of outlier points) "
        print "tuple format: [a_0,..., a_ndim, a_group, a_val]"
        print "Usage:\tpython gensinglecluster.py npts ndim kdim volperc [uh, sh, uo, so]"
        print "\tnpts: number of points per group (10 groups, 5 outlier, 5 normal)"
        print "\tkdim: dimensions of outlier cluster"
        print "\tvolperc: percentage volume of outlier cluster (default: 10%)"
        print "\tuh/uo: mean of normal and outlier a_val value (defaults: 10, 90)"
        print "\tsh/so: std of a_val values (defaults: 5, 5)"
        exit()


    npts, ndim, kdim = map(int, sys.argv[1:4])
    volperc = float(sys.argv[4])
    if len(sys.argv) > 5:
        uh, sh, uo, so = map(float, sys.argv[5:])
    else:
        uh, sh, uo, so = 10, 5, 90, 5

    # print schema
    outlier_box, super_box, schema,  pts = gen_points(npts, ndim, kdim, volperc, uh, sh, uo, so)
    print "\t".join(schema)
    for pt in pts:
        print '\t'.join(['%.4f'] * len(pt)) % tuple(pt)






"""
Generate dataset that is hard to greedy hill climb

If the search algorithm splits into 1/5s each time and then
the range x:1-20 y:1-20 should not affect avg, even though (2)
should eventually be detected.

Algorithm should also be confused because (1) is an obvious bad data
range

Data ranges
    x: 1 to 100
    y: 1 to 100
    z: 1 to 20
    val: -5 - 5

Bad data ranges
    1) x: 50 - 55, y: 50 - 55, val: 50
    2) x: 1 - 20, y: 1  - 10, val: 100
    3) x: 1 - 20, y: 10 - 20, val: -100


"""
import subprocess
import random
import sys
import os


if len(sys.argv) < 2:
    print "need argument of 1,2,3"
    exit()

harddata = int(sys.argv[1])
tablename = 'harddata_%d' % harddata


with file('/tmp/%s' % tablename, 'w') as f:

    ndim_cont = 10
    ndim_disc = 5
    # ndim_cont continuous dimensions.
    # ndim_disc discrete dimensions
    

    
    
    for z in xrange(20):
        for i in xrange(10000):
            x = random.random() * 100 
            y = random.random() * 100
            val = random.random() * 10 - 5

            if harddata == 1:
                if x >= 30 and x < 70 and y >= 30 and y <= 70:
                    if x >= 40 and x <= 60 and y >= 40 and y <= 60:
                        if z < 10:
                            val = 100 + random.random() * 10 - 5
                    else:
                        val = -100 + random.random() * 10 - 5

            if harddata == 2:        
                if (int(x / 20) + int(y / 20)) % 2:
                    val = 100 + random.random() * 10 - 5
                else:
                    val = -100 + random.random() * 10 - 5

            if harddata == 3:
                if z < 10:
                     if x >= 1 and x < 20:
                         if y >= 1 and y < 10:
                             val = 100 + random.random() * 10 - 5
                         elif y >= 7 and y < 20:
                             val = -(100. + random.random() * 10. - 5)
                     elif x >= 50 and x < 55 and y >= 50 and y < 55:
                         val = 50 + random.random() * 10 - 5

            
            print >> f, '%f,%f,%f,%f' % (x,y,z,val)


dbname = 'harddata'

try:
  os.system("psql -h localhost %s -c 'drop table %s'" % (dbname, tablename))
except Exception as e:
  print e

try:
  os.system("psql -h localhost %s -c 'create table %s (id serial, x float, y float, z float, v float)'" % (dbname, tablename))
except Exception as e:
  print e

os.system("psql -h localhost %s -c \"copy %s (x,y,z,v) from '/tmp/%s' with csv\"" % (dbname, tablename, tablename))



import json
import math
import pdb
import random
import numpy as np
import sys
import time
sys.path.extend(['.', '..'])

from itertools import chain, repeat, ifilter
from collections import defaultdict
from rtree.index import Index as RTree
from rtree.index import Property as RProp
from operator import mul, and_, or_

from ..util import rm_attr_from_domain, get_logger, instrument
from ..util.table import *
from ..util.misc import valid_number
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *

from adjgraph import AdjacencyGraph

_logger = get_logger()


class Merger(object):
    """
    Merges clusters

    transparently scales
    - cluster bounding boxes by table bounds
    - errors by min/max error
    """

    def __init__(self, **kwargs):
        self.min_clusters = 1
        self.influence = None
        self.learner = kwargs.get('learner', None)

        self.stats = defaultdict(lambda: [0, 0])
        self.adj_graph = None
        self.rtree = None
        self.base_clusters = []
        self.use_mtuples = kwargs.get('use_mtuples', True)
        self.use_cache = kwargs.get('use_cache', False)
        self.nmerged = 0

        # whether or not the partitions cover the entire space (can assume adjacency)
        # False if output from merger
        self.partitions_complete = kwargs.get('partitions_complete', True)

        self.DEBUG = kwargs.get('DEBUG', False)


        self.CACHENAME = './dbwipes.merger.cache'

        self.cache = None# bsddb3.hashopen('./dbwipes.merger.cache')
        
        self.set_params(**kwargs)

    def set_params(self, **kwargs):
        self.min_clusters = kwargs.get('min_clusters', self.min_clusters)
        # lambda cluster: influence_value_of(cluster)
        self.learner = kwargs.get('learner', self.learner)
        self.influence = self.learner.influence_cluster
        self.use_mtuples = kwargs.get('use_mtuples', self.use_mtuples)
        self.use_cache = kwargs.get('use_cache', self.use_cache)
        self.partitions_complete = kwargs.get('partitions_complete', self.partitions_complete)

        

    def setup_stats(self, clusters):
        """
        computes error bounds and the minimum volume of a 0-volume cluster

        """
        vols = np.array([c.volume for c in clusters if c.volume > 0])
        if len(vols) == 0:
            self.point_volume = 0.00001
        else:
            self.point_volume = vols.min() / 2.

    def close(self):
      """teardown code"""
      pass


    def setup_errors(self, clusters):
        return

    def get_states(self, merged, intersecting_clusters):
        @instrument
        def update_states(self, weight, global_states, efs, states, cards):
            if states is None:
                return

            thezip = zip(global_states, efs, states, cards)
            for idx, (gstate, ef, state, n) in enumerate(thezip):
                n = n * weight
                n = int(math.ceil(n))
                if n >= 1:
                    n = int(math.floor(n))
                else:
                    n = random.random() <= n and 1 or 0

                if n and state:
                    ustate = ef.update((state,), n)
                    if not gstate: 
                        global_states[idx] = ustate
                    else:
                        global_states[idx] = ef.update((ustate, gstate))


        bad_states = [None]*len(self.learner.bad_tables)
        good_states = [None]*len(self.learner.good_tables)
        bad_efs = self.learner.bad_err_funcs
        good_efs = self.learner.good_err_funcs
        for inter in intersecting_clusters:
            ibox = intersection_box(inter.bbox, merged.bbox)
            ivol = volume(ibox)
            if ivol < 0:
                continue
            weight = ivol / merged.volume
            update_states(self, weight, bad_states, bad_efs, inter.bad_states, inter.bad_cards)
            update_states(self, weight, good_states, good_efs, inter.good_states, inter.good_cards)

        return bad_states, good_states


    def influence_from_mtuples(self, merged, intersecting_clusters):
        """
        @deprecated!
        """
        bad_states, good_states = self.get_states(merged, intersecting_clusters)


        if not sum(map(bool, bad_states)):
            return None

        # now compute the influence using these m-tuple states
        @instrument
        def get_influences(self, efs, states, master_states,c ):
            infs = []
            for ef, state, mstate in zip(efs, states, master_states):
                if state:
                    # XXX
                    # XXX: HUGE HACK.  Takes count from state, assumes
                    # XXX: states is m-tuple of avg()
                    # XXX
                    influence = ef.recover(ef.remove(mstate, state)) 
                    if state[-1]**c:
                        influence = influence / (state[-1]**c)  
                        infs.append(inluencef)
            return infs

        bad_efs = self.learner.bad_err_funcs
        good_efs = self.learner.good_err_funcs
       
        bad_infs = get_influences(self, bad_efs, bad_states, self.learner.bad_states, self.learner.c)
        good_infs = map(abs, get_influences(self, good_efs, good_states, self.learner.good_states, 0) or [])
        if not bad_infs:
            return -1e10000000

        bad_inf = bad_infs and np.mean(bad_infs) or -1e100000000
        good_inf = good_infs and np.mean(good_infs) or 0
        l = self.learner.l
 
        return l * bad_inf - (1. - l) * good_inf

    @instrument
    def merge(self, cluster, neighbor, clusters):
      """
      @deprecated!
      """
      newbbox = bounding_box(cluster.bbox, neighbor.bbox)
      cidxs = self.get_intersection(newbbox)
      intersecting_clusters = [clusters[cidx] for cidx in cidxs]
      intersecting_clusters = filter(cluster.discretes_overlap, intersecting_clusters)

      merged = Cluster.merge(cluster, neighbor, intersecting_clusters, self.point_volume)
      if not merged or not merged.volume:
          return None
      
      if self.use_mtuples and cluster.discretes_same(neighbor):
          intersecting_clusters = filter(cluster.discretes_same, intersecting_clusters)
          merged.error = self.influence_from_mtuples(merged, intersecting_clusters)
      else:
          merged.error = self.influence(merged)
      return merged

    @instrument
    def dim_merge(self, cluster, dim, dec=None, inc=None, seen=None):
      bbox = [list(cluster.bbox[0]), list(cluster.bbox[1])]
      merged = cluster.clone()
      if dec is not None:
        if bbox[0][dim] <= dec:
          dec = None
        else:
          bbox[0][dim] = dec
      if inc is not None:
        if bbox[1][dim] >= inc:
          inc = None
        else:
          bbox[1][dim] = inc
      if dec is None and inc is None:
        return None
      merged.bbox = (tuple(bbox[0]), tuple(bbox[1]))
      if seen and merged.bound_hash in seen: 
        return None
      merged.rule = None

      start = time.time()
      merged.rule = merged.to_rule(
        self.learner.full_table,
        self.learner.cont_dists,
        self.learner.disc_dists
      )
      self.stats['merged.to_rule'][0] += time.time() - start
      self.stats['merged.to_rule'][1] += 1

      start = time.time()
      merged.error = self.influence(merged)
      self.stats['merged.influence'][0] += time.time() - start
      self.stats['merged.influence'][1] += 1

      merged.parents = [cluster]
      self.nmerged += 1
      if abs(merged.error) == float('inf'):
        return None
      return merged

    @instrument
    def disc_merge(self, cluster, dim, vals, seen=None):
      merged = cluster.clone()
      vals = set(vals)
      vals.update(merged.discretes.get(dim, ()))
      if len(merged.discretes[dim]) == len(vals):
        return None
      merged.discretes[dim] = vals
      if seen and merged.bound_hash in seen: 
        return None
      merged.rule = None
      start = time.time()
      merged.rule = merged.to_rule(
        self.learner.full_table,
        self.learner.cont_dists,
        self.learner.disc_dists
      )
      self.stats['merged.to_rule'][0] += time.time() - start
      self.stats['merged.to_rule'][1] += 1
      merged.error = self.influence(merged)
      merged.parents = [cluster]
      self.nmerged += 1

      if abs(merged.error) == float('inf'):
        return None
      return merged

    def dims_to_expand(self, cluster, seen=None, version=None):
      """
      Returns an iteratior of each dimension, direction, and values
      to expand along that dimension

      Args
        seen    set of clusters that have been already generated and
                can be skipped
      Yields (dim, dir, g)
            dim:  dimention
            dir:  direction of expansion ('inc', 'dec', 'disc')
            vals: values along this dimension and direction
      """
      def check(n):
        return not(
          (seen and n.bound_hash in seen) or 
          (n==cluster) or 
          cluster.same(n, epsilon=0.01) or 
          cluster.contains(n)
        )

      start = time.time()
      neighbors = self.adj_graph.neighbors(cluster, version=version)
      cost = time.time() - start
      self.stats['neighbor'][0] += cost
      self.stats['neighbor'][1] += 1
      _logger.debug("\t# neighbors: %d", len(neighbors))

      neighbors = filter(check, neighbors)
      if not neighbors: 
        return

      # gather all the directions and increments to expand along
      start = time.time()
      dim_to_incs = defaultdict(set)
      dim_to_decs = defaultdict(set)
      dim_to_discs = defaultdict(set)
      for n in neighbors:
        for dim, bound in enumerate(zip(*n.bbox)):
          minv, maxv = tuple(bound)
          if minv < cluster.bbox[0][dim]:
            dim_to_decs[dim].add(minv)
          if maxv > cluster.bbox[1][dim]:
            dim_to_incs[dim].add(maxv)

        disc_diffs = cluster.discrete_differences(n, epsilon=0.01)
        if disc_diffs:# and len(disc_diffs) <= 1:
          for dim, vals in disc_diffs.iteritems():
            dim_to_discs[dim].add(tuple(sorted(vals)))
            #dim_to_discs[dim].update(vals)
      cost = time.time() - start
      self.stats['gather'][0] += cost
      self.stats['gather'][1] += 1


      for dim, vals in dim_to_decs.iteritems():
        vals = filter(bool, sorted(vals, reverse=True))
        yield (dim, 'dec', vals)

      for dim, vals in dim_to_incs.iteritems():
        vals = filter(bool, sorted(vals))
        yield (dim, 'inc', vals)

      for dim, vals in dim_to_discs.iteritems():
        if sum(map(len, vals)) < 10:
          allvals = set()
          map(allvals.update, vals)
          vals = [set([v]) for v in allvals]
        else:
          vals = filter(bool, sorted(vals, key=lambda v: len(v)))
        yield (dim, 'disc', vals)


    def expand_candidates(self, cluster, seen=None, version=None):
      for dim, direction, vals in self.dims_to_expand(cluster, seen, version=version):
        if direction  == 'inc':
          generator = (
            self.dim_merge(cluster, dim, None, inc, seen) 
            for inc in vals
          )
        elif direction  == 'dec':
          generator = (
            self.dim_merge(cluster, dim, dec, None, seen) 
            for dec in vals
          )
        else:
          generator = (
            self.disc_merge(cluster, dim, disc)
            for disc in vals
          )

        generator = ifilter(bool, generator)
        yield (dim, direction, generator)



    def create_filterer(self, cluster, reasons=None):
      def filter_cluster(c, n=None):
        reason = None
        if c is None:
          reason = '.'
        if c.error == None:
          reason = 'N'
        if c.error == -1e10000000 or c.error == 1e1000000000:
          reason = 'e'
        if math.isnan(c.error):
          reason = '_'
        if c.error <= cluster.error:
          reason = '<%.4f'%c.error
        if n and c.error <= n.error:
          reason = '<n'
        ret = reason is not None

        if ret:
          reason = '!'
        if reasons is not None:
          reasons.append(reason)
        return ret
      return filter_cluster


    @instrument
    def expand(self, cluster, clusters):
      rms = set()
      while True:
        reasons = []
        expansions = self.expand_candidates(cluster, rms)
        filter_cluster = self.create_filterer(cluster, reasons)

        merged = cluster
        for dim, direction, g in expansions:
          for cand in g:
            if not filter_cluster(cand): break
            if cand.error > merged.error:
              merged = cand

        if not merged or merged == cluster:
          break

        _logger.debug('\tmerged:\t%s',merged)

        rms.update(merged.parents)
        cluster = merged
        self.adj_graph.insert(merged)

      _logger.debug('\n')
      return cluster, rms



    @instrument
    def cache_results(self, clusters_set, mergable_clusters):
        if not self.use_cache:
            return

        try:
            import bsddb as bsddb3
            self.cache =  bsddb3.hashopen(self.CACHENAME)
            myhash = str(hash(self.learner))
            c = str(self.learner.c)
            key = '%s:%s' % (myhash, c)

            clusters_set = [cluster.to_dict() for cluster in clusters_set]
            mergable_clusters = [cluster.to_dict() for cluster in mergable_clusters]
            self.cache[key] = json.dumps((clusters_set, mergable_clusters))

            cs_to_keys = json.loads(self.cache[myhash]) if myhash in self.cache else {}
            cs_to_keys[c] = key
            self.cache[myhash] = json.dumps(cs_to_keys)
            self.cache.close()
            _logger.debug("saved cache %f", self.learner.c)
        except:
            import traceback
            traceback.print_exc()



    @instrument
    def load_from_cache(self, clusters):
        """
        if there is cached info, load it and use the loaded data to
        1) replace structures like adj_graph and rtree
        2) initialize mergeable clusters, etc

        All state needs to be transient
        - adj_graph
        - rtree
        - clusters list
        - mergeable clusters
        - clusters_set
        
        """
        myhash = str(hash(self.learner))
        c = self.learner.c
        if self.use_cache:
          try:
            import bsddb as bsddb3
            self.cache =  bsddb3.hashopen(self.CACHENAME)
            if myhash not in self.cache:
                self.cache.close()
                raise RuntimeError("cache miss")
            cs_to_keys = json.loads(self.cache[myhash])
            cs_to_keys = dict([(float(k),v) for k,v in cs_to_keys.iteritems()])
            cs = [other_c for other_c in cs_to_keys if other_c >= c]
            key = str(cs_to_keys[min(cs)])
            matches = c == min(cs)

            clusters_set, mergable_clusters = json.loads(self.cache[key])
            clusters_set = set(map(Cluster.from_dict, clusters_set))
            mergable_clusters = map(Cluster.from_dict, mergable_clusters)

            also_mergable = []
            for cluster in clusters:
                useless = False
                for mc in mergable_clusters:
                    useless = useless or mc.contains(cluster)
                    if useless:
                        break
                if not useless:
                    self.adj_graph.insert(cluster)
                    also_mergable.append(cluster)
              
            mergable_clusters.extend(also_mergable)
            clusters_set.update(also_mergable)

            # fix up their error values!
            start = time.time()
            for cluster in chain(mergable_clusters, clusters_set):
              if self.use_mtuples:
                raise RuntimeError("disabled mtuples for now")
                cidxs = self.get_intersection(cluster.bbox)
                intersecting_clusters = [clusters[cidx] for cidx in cidxs]
                cluster.error = self.influence_from_mtuples(cluster, intersecting_clusters)
              else:
                cluster.error = self.influence(cluster)
            self.stats['load_fixup'][0] += time.time()-start
            self.stats['load_fixup'][1] += 1

            self.cache.close()
            alt_mergable = clusters
            if len(alt_mergable) > len(mergable_clusters):
                _logger.debug("loaded from cache\t%.4f", self.learner.c)
                mergable_clusters.sort(key=lambda mc: mc.error, reverse=True)
                return matches, clusters_set, mergable_clusters
            
            alt_mergable.sort(key=lambda mc: mc.error, reverse=True)
            return False, set(clusters), alt_mergable
          except:
            import traceback
            traceback.print_exc()


        clusters_set = set(clusters)
        mergable_clusters.sort(key=lambda mc: mc.error, reverse=True)
        return False, clusters_set, mergable_clusters


    @instrument
    def make_adjacency(self, clusters, partitions_complete=True):
        return AdjacencyGraph(
          clusters, 
          self.learner.full_table.domain, 
          self.learner.cont_dists
        )

        
    def __call__(self, clusters, **kwargs):
        if not clusters:
            return list(clusters)

        _logger.debug("merging %d clusters", len(clusters))

        self.set_params(**kwargs)
        self.setup_stats(clusters)

        # adj_graph is used to track adjacent partitions
        _logger.debug("building adj graph")
        self.adj_graph = self.make_adjacency(clusters, self.partitions_complete)



        # load state from cache
        can_stop, clusters_set, mergable_clusters = self.load_from_cache(clusters)
        if can_stop:
            return sorted(clusters_set, key=lambda c: c.error, reverse=True)

        _logger.debug("start merging!")
        while len(clusters_set) > self.min_clusters:

          cur_clusters = sorted(clusters_set, key=lambda c: c.error, reverse=True)

          _logger.debug("# mergable clusters: %d\tout of\t%d",
                  len(mergable_clusters),
                  len(cur_clusters))
          if not mergable_clusters:
              break


          merged_clusters, new_clusters = set(), set()
          seen = set()

          for cluster in mergable_clusters:
            if (cluster in merged_clusters or 
                cluster in new_clusters or cluster.bound_hash in seen):
                continue

            canskip = False
            for test in chain(new_clusters, mergable_clusters):
              if test == cluster: continue
              if test.contains(cluster, .01):
                #_logger.debug("skipped\n\t%s\n\t%s", str(cluster), str(test))
                canskip = True
                break
            if canskip:
              _logger.debug("skipped\n\t%s\n\t%s", str(cluster), str(test))
              continue

            merged, rms = self.expand(cluster, clusters) 
            if not merged or merged == cluster or len(filter(lambda c: c.contains(merged), cur_clusters)):
              seen.add(cluster.bound_hash)
              continue
            if not valid_number(merged.error):
              continue
            
            _logger.debug("%.4f\t%.4f\t-> %.4f",
                            merged.parents[0].error,
                            merged.parents[0].error,
                            merged.error)
            seen.update([c.bound_hash for c in merged.parents])
            seen.update([c.bound_hash for c in rms])


            if merged not in cur_clusters:
              new_clusters.add(merged)                    
            merged_clusters.update(rms)

          _logger.debug("merged %d\t%d new clusters\tout of %d",
                        len(merged_clusters),
                        len(new_clusters),
                        len(mergable_clusters))

          
          if not new_clusters:
              break


          map(self.adj_graph.remove, merged_clusters)
          map(self.adj_graph.insert, new_clusters)
          self.adj_graph.new_version()
          clusters_set.difference_update(merged_clusters)
          clusters_set.update(new_clusters)
          mergable_clusters = sorted(new_clusters, key=lambda c: c.error, reverse=True)

        clusters_set = filter_bad_clusters(clusters_set)
        self.cache_results(clusters_set, mergable_clusters)
        return sorted(clusters_set, key=lambda c: c.error, reverse=True)
            
    




import time
import pdb
import sys
import Orange
import orange
import heapq
sys.path.extend(['.', '..'])

from itertools import chain
from scorpionsql.errfunc import *


from ..learners.cn2sd.rule import fill_in_rules
from ..learners.cn2sd.refiner import *
from ..bottomup.bounding_box import *
from ..bottomup.cluster import *
from ..util import *

from basic import Basic
from merger import Merger

_logger = get_logger()


class Blah(object):
  def __init__(
      self, attrs, group, 
      bds, bcs, gds, gcs, 
      maxinf, mr, grouper):

    self.attrs = attrs
    self.grouper = grouper
    self.group = group
    self.inf_state = (bds, bcs, gds, gcs)
    self.mr = mr
    self.c_range = list(mr.c_range)
    self._rule = None

    c = mr.c
    l = mr.l

    good_infs = [abs(gd) for gd,gc in zip(gds, gcs) if gc]
    #self.maxg = max(good_infs) if good_infs else 0
    self.maxg = np.percentile(good_infs, 85) if good_infs else 0
    self.good_inf = (1. - l) * self.maxg

    self.bad_inf_func = self.create_bad_inf_func()
    self.inf_func = self.create_inf_func()

    self.best_tuple_inf = l * maxinf

    self.bad_inf = self.bad_inf_func(c)
    self.inf = self.inf_func(c)
    self.maxinf = maxinf
    self.npts = sum(bcs) + sum(gcs)


    # @deprecated
    self.best_inf = self.bad_inf
    self.good_skip = False # subpredicates can skip computing good_inf

  def create_bad_inf_func(self):
    """
    copy and pasted from Basic.create_inf_func
    """
    bds, bcs = [], []
    for idx in xrange(len(self.inf_state[0])):
      bd, bc = self.inf_state[0][idx], self.inf_state[1][idx]
      if valid_number(bd) and valid_number(bc):
        bds.append(bd)
        bcs.append(bc)
      else:
        bds.append(0)
        bcs.append(0)

    def f(c, l=None):
      l = l if l is not None else self.mr.l
      return compute_influence(l, compute_bad_score(bds, bcs, c), 0)
    return f

  def create_inf_func(self):
    l = self.mr.l
    f = lambda c: compute_influence(l, self.bad_inf_func(c, 1.), self.maxg)
    return f

  def dominated_by(self, o, c_range=None):
    if not c_range: c_range = self.c_range
    mybads = map(self.bad_inf_func, c_range)
    oinfs = map(o.inf_func, c_range)
    if r_lte(mybads, oinfs):
      if self.best_tuple_inf <= min(oinfs):
        return True
    return False


  def clone(self):
    args = [self.attrs, self.group] 
    args.extend(self.inf_state)
    args.extend((self.maxinf, self.mr, self.grouper))
    ret = Blah(*args)
    ret._rule = self._rule.clone()
    return ret


  def __rule__(self):
    if self._rule: return self._rule

    conds = []
    for attr, gid in zip(self.attrs, self.group):
      if attr.var_type ==  Orange.feature.Type.Discrete:
        vals = [orange.Value(attr, v) for v in self.grouper.id2vals[attr][gid]]
        conds.append(
          orange.ValueFilter_discrete(
              position = self.grouper.data.domain.index(attr),
              values = vals
          )
        )
      else:
        vals = self.grouper.id2vals[attr][gid]
        minv, maxv = vals[0], vals[1]
        conds.append(
          Orange.data.filter.ValueFilterContinuous(
            oper=orange.ValueFilter.Between,
            position = self.grouper.data.domain.index(attr),
            min=minv,
            max=maxv
          )
        )
    self._rule = SDRule(self.grouper.data, None, conds, None)
    self._rule.quality = self.inf
    self._rule.inf_state = self.inf_state

    return self._rule
  rule = property(__rule__)


  def __str__(self):
    args = (
      self.inf, 
      sum(self.inf_state[1]), 
      sum(self.inf_state[3]), 
      self.bad_inf, self.good_inf, 
      self.rule
    ) 
    return 'inf %.4f\tpts %d/%d\tr/g %.4f - %.4f\t%s' % args

  def __hash__(self):
      return hash(str(self.group) + str(self.attrs))

  def __eq__(self, o):
      return hash(self) == hash(o)


  def __cmp__(self, o):
      if self.inf > o.inf:
          return 1
      if self.inf == o.inf:
          return 0
      return -1




class Grouper(object):
    def __init__(self, table, mr):
        self.data = table
        self.mr = mr
        self.ddists = None
        self.bdists = None
        self.gbids = {}
        self.id2vals = {}
        self.gbfuncs = {}
        self.mappers = {}
        self.setup_functions()

    def setup_functions(self):
        domain = self.data.domain
        ddists = Orange.statistics.distribution.Domain(self.data)
        self.ddists = ddists
        bdists = Orange.statistics.basic.Domain(self.data)
        self.bdists = bdists
        gbfuncs = {}
        gbids = {}
        id2vals = {}

        for idx, attr in enumerate(self.data.domain):
          if attr.name not in self.mr.cols:
            continue
          if attr.var_type == Orange.feature.Type.Discrete:
            groups = self.create_discrete_groups(attr, idx, ddists[idx].keys())
            mapper = {}
            for gidx, group in enumerate(groups):
                for val in group:
                    mapper[val] = gidx
            self.mappers[attr] = mapper

            f = lambda v: int(self.mappers[v.variable].get(v.value, len(groups)))
            n = len(groups)
            ranges = groups
          else:
            dist = bdists[idx]
            maxv, minv = dist.max, dist.min
            if maxv == minv:
              f = lambda v: 0
              n = 1
              ranges = [(minv, maxv)]
            else:
              def make_func(minv, block):
                def f(v):
                  return int(math.ceil((v-minv)/block)) 
                return f
              block = (maxv - minv) / float(self.mr.granularity)
              n = self.mr.granularity #+1
              f = make_func(minv, block)
              ranges = [(maxv - (i+1)*block, maxv - (i)*block) for i in xrange(n-1,-1,-1)]
              ranges[0] = [-float('infinity'), ranges[0][1]]
              ranges[-1] = [ranges[-1][0], float('infinity')]
              print attr, block, minv, maxv

          gbfuncs[attr] = f
          gbids[attr] = n
          id2vals[attr] = dict(enumerate(ranges))

        self.gbfuncs = gbfuncs
        self.gbids = gbids
        self.id2vals = id2vals

    def create_discrete_groups(self, attr, pos, vals):
        return [(val,) for val in sorted(vals)]
        if len(vals) == 1:
            return (vals,)

        rule = SDRule(self.data, None, [orange.ValueFilter_discrete(
            position = pos, 
            values = [orange.Value(attr,v) for v in vals]
        )], None)
        ro = RuleObj(rule, self.mr)

        if not self.mr.prune_rule(ro):
            return (vals,)

        ret = []
        for newvals in block_iter(vals, 2):
            ret.extend(self.create_discrete_groups(attr, pos, newvals))
        return ret


    def _get_infs(self, all_table_rows, err_funcs, g, bmaxinf):
      """
      Args:
        all_table_rows: list, each item is a dict of
                        group -> rows in that group (e.g., predicate)
        g: the group
        bmaxinf: want max inf of individual rows
      """
      ret = []
      counts = []
      maxinf = float('-inf')
      iterator = zip(enumerate(err_funcs), all_table_rows)
      for (idx, ef), table_rows in iterator:
        rows = table_rows.get(g, [])
        if not rows:
          continue
        
        if bmaxinf:
          for row in rows:
            curinf = self.influence_tuple(row, ef)
            maxinf = max(maxinf, curinf) 

        ret.append(ef(rows))
        counts.append(len(rows))
      return ret, counts, maxinf

    def influence_tuple(self, row, ef):
      if row[self.mr.SCORE_ID].value == float('-inf'):
        influence = ef((row,))
        row[self.mr.SCORE_ID] = influence
      return row[self.mr.SCORE_ID].value

    def groups_by_attrs(self, attrs, valid_groups, table):
      """scan table once and group tuples by their respective groups"""
      groups = defaultdict(list)
      for row in table:
        group = tuple([self.gbfuncs[attr](row[attr]) for attr in attrs])
        if group in valid_groups:
          groups[group].append(row)
      return groups


    def __call__(self, attrs, valid_groups):
      valid_groups = set(valid_groups)
      bad_table_rows = []
      good_table_rows = []
      for table in self.mr.bad_tables:
        bad_table_rows.append(self.groups_by_attrs(attrs, valid_groups, table))
      for table in self.mr.good_tables:
        good_table_rows.append(self.groups_by_attrs(attrs, valid_groups, table))

      for g in valid_groups:
        # compute the influence_state for each group (cube)
        bds, bcs, maxinf = self._get_infs(bad_table_rows, self.mr.bad_err_funcs, g, True)
        gds, gcs, _ = self._get_infs(good_table_rows, self.mr.good_err_funcs, g, False)
        if not bcs:
          continue
        yield Blah(attrs, g, bds, bcs, gds, gcs, maxinf, self.mr, self)




    def initial_groups(self):
      keys = sorted(self.gbids.keys(), key=lambda a: a.name)
      for attr in keys:
        n = self.gbids[attr]
        yield (attr,), ((i,) for i in xrange(n))




    def merge_groups(self, prev_groups):
      """
      prev_groups: attributes -> groups
      attributes are sorted
      group: attr -> idx
      """
      start = time.time()
      attrs_list = sorted(prev_groups.keys())
      for a_idx, attrs1 in enumerate(attrs_list):
        sattrs1 = set(attrs1)
        pgroup1 = [dict(zip(attrs1, g)) for g in prev_groups[attrs1]]
        for attrs2 in attrs_list[a_idx+1:]:
          pgroup2 = [dict(zip(attrs2, g)) for g in prev_groups[attrs2]]

          sattrs2 = set(attrs2)
          merged_attrs = tuple(sorted(sattrs1.union(sattrs2)))
          if len(merged_attrs) != len(sattrs1)+1:
              continue
          intersecting = tuple(sorted(sattrs1.intersection(sattrs2)))

          unique1 = tuple(sattrs1.difference(sattrs2))[0]
          unique2 = tuple(sattrs2.difference(sattrs1))[0]
          idx1 = merged_attrs.index(unique1)
          idx2 = merged_attrs.index(unique2)

          diff = time.time() - self.mr.start
          if diff >= self.mr.max_wait:
            _logger.debug("wait %d > %d exceeded" % (diff, self.mr.max_wait))
            return

          yield (merged_attrs, self.fulljoin(intersecting, merged_attrs, idx1, idx2, pgroup1, pgroup2))

    def fulljoin(self, inter, union, idx1, idx2, groups1, groups2):
      def make_key(group):
        ret = [group.get(k, None) for k in union]
        ret[idx1] = ret[idx2] = None
        return tuple(ret)

      matches1 = defaultdict(list)
      matches2 = defaultdict(list)
      try:
        for g1 in groups1:
            matches1[make_key(g1)].append(g1[union[idx1]])
        for g2 in groups2:
            matches2[make_key(g2)].append(g2[union[idx2]])
      except:
        pdb.set_trace()

      seen = set()

      for key in matches1.keys():
        if key not in matches2:
          continue

        for v1 in matches1[key]:
          for v2 in matches2[key]:
            newg = list(key)
            newg[idx1] = v1
            newg[idx2] = v2
            newg = tuple(newg)
            if newg in seen:
              continue
            seen.add(newg)
      return seen

             
